# -*- coding: utf-8 -*-
# pylint: disable=r0904, c0415, c0413, r0913, w0212
""" unittests for dkb_robo """
import sys
import os
from datetime import date
import unittest
import logging
import json
from unittest.mock import patch, Mock, mock_open
from bs4 import BeautifulSoup
from mechanicalsoup import LinkNotFoundError
sys.path.insert(0, '.')
sys.path.insert(0, '..')
from dkb_robo.legacy import Wrapper


def json_load(fname):
    """ simple json load """

    with open(fname, 'r', encoding='utf8') as myfile:
        data_dic = json.load(myfile)

    return data_dic


def read_file(fname):
    """ read file into string """
    with open(fname, "rb") as myfile:
        data = myfile.read()

    return data


def cnt_list(value):
    """ customized function return just the number if entries in input list """
    return len(value)


def my_side_effect(*args):
    """ my sideeffect funtion """
    return [200, args[1], [args[4]]]


@patch('dkb_robo.legacy.Wrapper.dkb_br')
class TestDKBRobo(unittest.TestCase):
    """ test class """

    maxDiff = None

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.wrapper = Wrapper()

    def test_001_get_cc_limit(self, mock_browser):
        """ test Legacy._get_credit_limits() method """
        html = read_file(self.dir_path + '/mocks/konto-kreditkarten-limits.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {'1111********1111': 100.00,
                    '1111********1112': 2000.00,
                    'DE01 1111 1111 1111 1111 11': 1000.00,
                    'DE02 1111 1111 1111 1111 12': 2000.00}

        self.assertEqual(e_result, self.wrapper.get_credit_limits())

    def test_002_get_cc_limit(self, mock_browser):
        """ test legacy.get_credit_limits() triggers exceptions """
        html = read_file(self.dir_path + '/mocks/konto-kreditkarten-limits-exception.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {'DE01 1111 1111 1111 1111 11': 1000.00, 'DE02 1111 1111 1111 1111 12': 2000.00}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(e_result, self.wrapper.get_credit_limits())
        self.assertIn("ERROR:dkb_robo.legacy:Wrapper.get_credit_limits() get credit card limits: 'NoneType' object has no attribute 'find'\n", lcm.output)

    def test_003_get_cc_limit(self, mock_browser):
        """ test legacy.get_credit_limits() no limits """
        # html = read_file(self.dir_path + '/mocks/konto-kreditkarten-limits-exception.html')
        html = '<html><body>fooo</body></html>'
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {}
        self.assertEqual(e_result, self.wrapper.get_credit_limits())

    def test_004_get_exo_single(self, mock_browser):
        """ test legacy.get_exemption_order() method for a single exemption order """
        html = read_file(self.dir_path + '/mocks/freistellungsauftrag.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {1: {'available': 1000.0, 'amount': 1000.0, 'used': 0.0, 'description': 'Gemeinsam mit Firstname Familyname', 'validity': '01.01.2016 unbefristet'}}
        self.assertEqual(self.wrapper.get_exemption_order(), e_result)

    def test_005_get_exo_single_nobr(self, mock_browser):
        """ test legacy.get_exemption_order() method for a single exemption order without line-breaks"""
        html = read_file(self.dir_path + '/mocks/freistellungsauftrag-nobr.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {1: {'available': 1000.0, 'amount': 1000.0, 'used': 0.0, 'description': 'Gemeinsam mit Firstname Familyname', 'validity': '01.01.2016 unbefristet'}}
        self.assertEqual(self.wrapper.get_exemption_order(), e_result)

    def test_006_get_exo_multiple(self, mock_browser):
        """ test legacy.get_exemption_order() method for a multiple exemption orders """
        html = read_file(self.dir_path + '/mocks/freistellungsauftrag-multiple.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {1: {'available': 1000.0,
                        'amount': 1000.0,
                        'used': 0.0,
                        'description': 'Gemeinsam mit Firstname1 Familyname1',
                        'validity': '01.01.2016 unbefristet'},
                    2: {'available': 2000.0,
                        'amount': 2000.0,
                        'used': 0.0,
                        'description': 'Gemeinsam mit Firstname2 Familyname2',
                        'validity': '02.01.2016 unbefristet'}}

        self.assertEqual(self.wrapper.get_exemption_order(), e_result)

    def test_007_get_exo_single(self, mock_browser):
        """ test legacy.get_exemption_order() method try throws exception """
        html = read_file(self.dir_path + '/mocks/freistellungsauftrag-indexerror.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {1: {}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(self.wrapper.get_exemption_order(), e_result)
        self.assertIn('ERROR:dkb_robo.legacy:Wrapper.get_exemption_order(): list index out of range\n', lcm.output)

    def test_008_new_instance(self, _unused):
        """ test legacy._new_instance() method """
        self.assertIn('mechanicalsoup.stateful_browser.StatefulBrowser object at', str(self.wrapper._new_instance()))

    def test_009_new_instance(self, _unused):
        """ test legacy._new_instance() method with proxies """
        self.wrapper.proxies = 'proxies'
        self.assertIn('mechanicalsoup.stateful_browser.StatefulBrowser object at', str(self.wrapper._new_instance()))
        self.assertEqual('proxies', self.wrapper.dkb_br.session.proxies)

    def test_010_new_instance(self, _unused):
        """ test legacy._new_instance() method with clientcookies """
        cookieobj = Mock()
        cookieobj.value = 'value'
        cookieobj.name = 'name'
        self.assertIn('mechanicalsoup.stateful_browser.StatefulBrowser object at', str(self.wrapper._new_instance([cookieobj])))

    def test_011_get_points(self, mock_browser):
        """ test legacy.get_points() method """
        html = read_file(self.dir_path + '/mocks/dkb_punkte.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {'DKB-Punkte': 100000, 'davon verfallen zum  31.12.2017': 90000}
        self.assertEqual(self.wrapper.get_points(), e_result)

    def test_012_get_so_multiple(self, mock_browser):
        """ test legacy._get_standing_orders() method """
        html = read_file(self.dir_path + '/mocks/dauerauftraege.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = [{'amount': 100.0, 'interval': '1. monatlich 01.03.2017', 'recipient': 'RECPIPIENT-1', 'purpose': 'KV 1234567890'},
                    {'amount': 200.0, 'interval': '1. monatlich geloescht', 'recipient': 'RECPIPIENT-2', 'purpose': 'KV 0987654321'}]
        self.assertEqual(self.wrapper.get_standing_orders(), e_result)

    @patch('dkb_robo.legacy.Wrapper._parse_overview')
    @patch('dkb_robo.legacy.Wrapper._get_financial_statement')
    @patch('dkb_robo.legacy.Wrapper._login_confirm')
    @patch('dkb_robo.legacy.Wrapper._ctan_check')
    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_013_login(self, mock_instance, mock_ctan, mock_confirm, mock_fs, mock_pov, mock_browser):
        """ test legacy._login() method - no confirmation """
        html = """
                <h1>Anmeldung bestätigen</h1>
                <div id="lastLoginContainer" class="lastLogin deviceFloatRight ">
                        Letzte Anmeldung:
                        01.03.2017, 01:00 Uhr
                </div>
               """
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_ctan.return_value = False
        mock_confirm.return_value = False
        mock_instance.return_value = mock_browser
        mock_pov.return_value = '_parse_overview'
        mock_fs.return_value = 'mock_fs'
        self.assertEqual(({}, '01.03.2017, 01:00 Uhr'), self.wrapper.login())
        self.assertTrue(mock_confirm.called)
        self.assertFalse(mock_ctan.called)
        self.assertFalse(mock_fs.called)
        self.assertFalse(mock_pov.called)

    @patch('dkb_robo.legacy.Wrapper._parse_overview')
    @patch('dkb_robo.legacy.Wrapper._get_financial_statement')
    @patch('dkb_robo.legacy.Wrapper._login_confirm')
    @patch('dkb_robo.legacy.Wrapper._ctan_check')
    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_014_login(self, mock_instance, mock_ctan, mock_confirm, mock_fs, mock_pov, mock_browser):
        """ test legacy.login() method - confirmation """
        html = """
                <h1>Anmeldung bestätigen</h1>
                <div id="lastLoginContainer" class="lastLogin deviceFloatRight ">
                        Letzte Anmeldung:
                        01.03.2017, 01:00 Uhr
                </div>
               """
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_ctan.return_value = False
        mock_confirm.return_value = True
        mock_instance.return_value = mock_browser
        mock_pov.return_value = '_parse_overview'
        mock_fs.return_value = 'mock_fs'
        self.assertEqual(('_parse_overview', '01.03.2017, 01:00 Uhr'), self.wrapper.login())
        self.assertTrue(mock_confirm.called)
        self.assertFalse(mock_ctan.called)
        self.assertTrue(mock_fs.called)
        self.assertTrue(mock_pov.called)

    @patch('dkb_robo.legacy.Wrapper._parse_overview')
    @patch('dkb_robo.legacy.Wrapper._get_financial_statement')
    @patch('dkb_robo.legacy.Wrapper._login_confirm')
    @patch('dkb_robo.legacy.Wrapper._ctan_check')
    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_015_login(self, mock_instance, mock_ctan, mock_confirm, mock_fs, mock_pov, mock_browser):
        """ test legacy._login() method - no confirmation """
        html = """
                <h1>Anmeldung bestätigen</h1>
                <div id="lastLoginContainer" class="lastLogin deviceFloatRight ">
                        Letzte Anmeldung:
                        01.03.2017, 01:00 Uhr
                </div>
               """
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_ctan.return_value = False
        mock_confirm.return_value = False
        mock_instance.return_value = mock_browser
        mock_pov.return_value = '_parse_overview'
        mock_fs.return_value = 'mock_fs'
        self.wrapper.tan_insert = True
        self.assertEqual(({}, '01.03.2017, 01:00 Uhr'), self.wrapper.login())
        self.assertFalse(mock_confirm.called)
        self.assertTrue(mock_ctan.called)
        self.assertFalse(mock_fs.called)
        self.assertFalse(mock_pov.called)

    @patch('dkb_robo.legacy.Wrapper._parse_overview')
    @patch('dkb_robo.legacy.Wrapper._get_financial_statement')
    @patch('dkb_robo.legacy.Wrapper._login_confirm')
    @patch('dkb_robo.legacy.Wrapper._ctan_check')
    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_016_login(self, mock_instance, mock_ctan, mock_confirm, mock_fs, mock_pov, mock_browser):
        """ test legacy._login() method - confirmation """
        html = """
                <h1>Anmeldung bestätigen</h1>
                <div id="lastLoginContainer" class="lastLogin deviceFloatRight ">
                        Letzte Anmeldung:
                        01.03.2017, 01:00 Uhr
                </div>
               """
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_ctan.return_value = True
        mock_confirm.return_value = False
        mock_instance.return_value = mock_browser
        mock_pov.return_value = '_parse_overview'
        mock_fs.return_value = 'mock_fs'
        self.wrapper.tan_insert = True
        self.assertEqual(('_parse_overview', '01.03.2017, 01:00 Uhr'), self.wrapper.login())
        self.assertFalse(mock_confirm.called)
        self.assertTrue(mock_ctan.called)
        self.assertTrue(mock_fs.called)
        self.assertTrue(mock_pov.called)

    @patch('dkb_robo.legacy.Wrapper._parse_overview')
    @patch('dkb_robo.legacy.Wrapper._get_financial_statement')
    @patch('dkb_robo.legacy.Wrapper._login_confirm')
    @patch('dkb_robo.legacy.Wrapper._ctan_check')
    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_017_login(self, mock_instance, mock_ctan, mock_confirm, mock_fs, mock_pov, mock_browser):
        """ test legacy._login() login failed """
        html = """
                <div id="lastLoginContainer" class="clearfix module text errorMessage">foo</div>
               """
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_ctan.return_value = True
        mock_confirm.return_value = False
        mock_instance.return_value = mock_browser
        mock_pov.return_value = '_parse_overview'
        mock_fs.return_value = 'mock_fs'
        self.wrapper.tan_insert = True
        with self.assertRaises(Exception) as err:
            self.assertEqual(self.wrapper.login(), None)
        self.assertEqual('Login failed', str(err.exception))
        self.assertFalse(mock_confirm.called)
        self.assertFalse(mock_ctan.called)
        self.assertFalse(mock_fs.called)
        self.assertFalse(mock_pov.called)

    @patch('dkb_robo.legacy.Wrapper._parse_overview')
    @patch('dkb_robo.legacy.Wrapper._get_financial_statement')
    @patch('dkb_robo.legacy.Wrapper._login_confirm')
    @patch('dkb_robo.legacy.Wrapper._ctan_check')
    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_018_login(self, mock_instance, mock_ctan, mock_confirm, mock_fs, mock_pov, mock_browser):
        """ test legacy._login() login failed """
        mock_browser.select_form.side_effect = LinkNotFoundError
        mock_ctan.return_value = True
        mock_confirm.return_value = False
        mock_instance.return_value = mock_browser
        mock_pov.return_value = '_parse_overview'
        mock_fs.return_value = 'mock_fs'
        self.wrapper.tan_insert = True
        with self.assertRaises(Exception) as err:
            self.assertEqual(self.wrapper.login(), None)
        self.assertEqual('Login failed: LinkNotFoundError', str(err.exception))
        self.assertFalse(mock_confirm.called)
        self.assertFalse(mock_ctan.called)
        self.assertFalse(mock_fs.called)
        self.assertFalse(mock_pov.called)

    @patch('dkb_robo.legacy.Wrapper._parse_overview')
    @patch('dkb_robo.legacy.Wrapper._get_financial_statement')
    @patch('dkb_robo.legacy.Wrapper._login_confirm')
    @patch('dkb_robo.legacy.Wrapper._ctan_check')
    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_019_login(self, mock_instance, mock_ctan, mock_confirm, mock_fs, mock_pov, mock_browser):
        """ test legacy._login() method - notice form """
        html = """
                <h1>Anmeldung bestätigen</h1>
                <form id="genericNoticeForm">foo</form>
                <div id="lastLoginContainer" class="lastLogin deviceFloatRight ">
                        Letzte Anmeldung:
                        01.03.2017, 01:00 Uhr
                </div>
               """
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_ctan.return_value = True
        mock_confirm.return_value = False
        mock_instance.return_value = mock_browser
        mock_pov.return_value = '_parse_overview'
        mock_fs.return_value = 'mock_fs'
        self.wrapper.tan_insert = True
        self.assertEqual(('_parse_overview', '01.03.2017, 01:00 Uhr'), self.wrapper.login())
        self.assertFalse(mock_confirm.called)
        self.assertTrue(mock_ctan.called)
        self.assertTrue(mock_fs.called)
        self.assertTrue(mock_pov.called)

    def test_020_parse_overview(self, _unused):
        """ test DKBRobo._parse_overview() method """
        html = read_file(self.dir_path + '/mocks/finanzstatus.html')
        e_result = {0: {'account': 'XY99 1111 1111 0000 1111 99',
                        'amount': 1367.82,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=0&group=0',
                        'name': 'hauptkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=0',
                        'type': 'account'},
                    1: {'account': '9999********1111',
                        'amount': 9613.31,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=0',
                        'name': 'first visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=0',
                        'type': 'creditcard'},
                    2: {'account': '9999********8888',
                        'amount': -260.42,
                        'date': '26.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=0',
                        'name': 'MilesnMoreMaster',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=0',
                        'type': 'creditcard'},
                    3: {'account': '9999********2222',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=0&group=1',
                        'name': 'second visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=1',
                        'type': 'creditcard'},
                    4: {'account': 'XY99 1111 1111 2155 2788 99',
                        'amount': 588.37,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=1',
                        'name': 'zweitkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=1',
                        'type': 'account'},
                    5: {'account': '9999********3333',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=1',
                        'name': '3rd visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=1',
                        'type': 'creditcard'},
                    6: {'account': 'XY99 3333 1111 0000 3333 99',
                        'amount': -334.34,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=3&group=1',
                        'name': '3rd acc',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=3&group=1',
                        'type': 'account'}}

        self.assertEqual(self.wrapper._parse_overview(BeautifulSoup(html, 'html5lib')), e_result)

    def test_021_parse_overview(self, _unused):
        """ test DKBRobo._parse_overview() method """
        html = read_file(self.dir_path + '/mocks/finanzstatus-error1.html')
        e_result = {0: {'account': 'XY99 1111 1111 0000 1111 99',
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=0&group=0',
                        'name': 'hauptkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=0',
                        'type': 'account'},
                    1: {'account': '9999********1111',
                        'amount': 9613.31,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=0',
                        'name': 'first visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=0',
                        'type': 'creditcard'},
                    2: {'account': '9999********8888',
                        'amount': -260.42,
                        'date': '26.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=0',
                        'name': 'MilesnMoreMaster',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=0',
                        'type': 'creditcard'},
                    3: {'account': '9999********2222',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=0&group=1',
                        'name': 'second visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=1',
                        'type': 'creditcard'},
                    4: {'account': 'XY99 1111 1111 2155 2788 99',
                        'amount': 588.37,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=1',
                        'name': 'zweitkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=1',
                        'type': 'account'},
                    5: {'account': '9999********3333',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=1',
                        'name': '3rd visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=1',
                        'type': 'creditcard'},
                    6: {'account': 'XY99 3333 1111 0000 3333 99',
                        'amount': -334.34,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=3&group=1',
                        'name': '3rd acc',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=3&group=1',
                        'type': 'account'}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(self.wrapper._parse_overview(BeautifulSoup(html, 'html5lib')), e_result)
        self.assertIn("ERROR:dkb_robo.legacy:Wrapper._parse_overview() convert amount: could not convert string to float: 'aaa'\n", lcm.output)

    def test_022_parse_overview(self, _unused):
        """ test DKBRobo._parse_overview() exception detail link"""
        html = read_file(self.dir_path + '/mocks/finanzstatus-error2.html')
        e_result = {0: {'account': 'XY99 1111 1111 0000 1111 99',
                        'amount': 1367.82,
                        'date': '27.04.2018',
                        'name': 'hauptkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=0',
                        'type': 'account'},
                    1: {'account': '9999********1111',
                        'amount': 9613.31,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=0',
                        'name': 'first visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=0',
                        'type': 'creditcard'},
                    2: {'account': '9999********8888',
                        'amount': -260.42,
                        'date': '26.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=0',
                        'name': 'MilesnMoreMaster',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=0',
                        'type': 'creditcard'},
                    3: {'account': '9999********2222',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=0&group=1',
                        'name': 'second visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=1',
                        'type': 'creditcard'},
                    4: {'account': 'XY99 1111 1111 2155 2788 99',
                        'amount': 588.37,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=1',
                        'name': 'zweitkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=1',
                        'type': 'account'},
                    5: {'account': '9999********3333',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=1',
                        'name': '3rd visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=1',
                        'type': 'creditcard'},
                    6: {'account': 'XY99 3333 1111 0000 3333 99',
                        'amount': -334.34,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=3&group=1',
                        'name': '3rd acc',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=3&group=1',
                        'type': 'account'}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(self.wrapper._parse_overview(BeautifulSoup(html, 'html5lib')), e_result)
        self.assertIn("ERROR:dkb_robo.legacy:Wrapper._parse_overview() get link: 'NoneType' object is not subscriptable\n", lcm.output)

    def test_023_parse_overview(self, _unused):
        """ test DKBRobo._parse_overview() exception depot """
        html = read_file(self.dir_path + '/mocks/finanzstatus-error3.html')
        e_result = {0: {'account': 'XY99 1111 1111 0000 1111 99',
                        'amount': 1367.82,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=0&group=0',
                        'name': 'hauptkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=0',
                        'type': 'account'},
                    1: {'account': '9999********1111',
                        'amount': 9613.31,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=0',
                        'name': 'first visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=0',
                        'type': 'creditcard'},
                    2: {'account': '9999********8888',
                        'amount': -260.42,
                        'date': '26.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=0',
                        'name': 'MilesnMoreMaster',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=0',
                        'type': 'creditcard'},
                    3: {'account': '9999********2222',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=0&group=1',
                        'name': 'second visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=0&group=1',
                        'type': 'creditcard'},
                    4: {'account': 'XY99 1111 1111 2155 2788 99',
                        'amount': 588.37,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=1&group=1',
                        'name': 'zweitkonto',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=1&group=1',
                        'type': 'account'},
                    5: {'account': '9999********3333',
                        'amount': 515.52,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=2&group=1',
                        'name': '3rd visa',
                        'transactions': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=paymentTransaction&row=2&group=1',
                        'type': 'creditcard'},
                    6: {'account': 'XY99 3333 1111 0000 3333 99',
                        'amount': -334.34,
                        'date': '27.04.2018',
                        'details': 'https://www.ib.dkb.de/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=details&row=3&group=1',
                        'name': '3rd acc',
                        'type': 'depot'}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(e_result, self.wrapper._parse_overview(BeautifulSoup(html, 'html5lib')))
        self.assertIn("ERROR:dkb_robo.legacy:Wrapper._parse_overview() parse depot: 'NoneType' object is not subscriptable\n", lcm.output)

    def test_024_parse_overview_mbank(self, _unused):
        """ test DKBRobo._parse_overview() method for accounts from other banks"""
        html = read_file(self.dir_path + '/mocks/finanzstatus-mbank.html')
        e_result = {0: {'account': '1111********1111',
                        'name': 'credit-card-1',
                        'transactions': 'https://www.ib.dkb.de/tcc-1',
                        'amount': 1000.0,
                        'details': 'https://www.ib.dkb.de/dcc-1',
                        'date': '01.03.2017',
                        'type': 'creditcard'},
                    1: {'account': '1111********1112',
                        'name': 'credit-card-2',
                        'transactions': 'https://www.ib.dkb.de/tcc-2',
                        'amount': 2000.0,
                        'details': 'https://www.ib.dkb.de/dcc-2',
                        'date': '02.03.2017',
                        'type': 'creditcard'},
                    2: {'account': 'DE11 1111 1111 1111 1111 11',
                        'name': 'checking-account-1',
                        'transactions': 'https://www.ib.dkb.de/tac-1',
                        'amount': 1000.0,
                        'details': 'https://www.ib.dkb.de/banking/dac-1',
                        'date': '03.03.2017',
                        'type': 'account'},
                    3: {'account': 'DE11 1111 1111 1111 1111 12',
                        'name': 'checking-account-2',
                        'transactions': 'https://www.ib.dkb.de/tac-2',
                        'amount': 2000.0,
                        'details': 'https://www.ib.dkb.de/banking/dac-2',
                        'date': '04.03.2017',
                        'type': 'account'},
                    4: {'account': '1111111',
                        'name': 'Depot-1',
                        'transactions': 'https://www.ib.dkb.de/tdepot-1',
                        'amount': 5000.0,
                        'details': 'https://www.ib.dkb.de/ddepot-1',
                        'date': '06.03.2017',
                        'type': 'depot'},
                    5: {'account': '1111112',
                        'name': 'Depot-2',
                        'transactions': 'https://www.ib.dkb.de/tdepot-2',
                        'amount': 6000.0,
                        'details': 'https://www.ib.dkb.de/ddepot-2',
                        'date': '06.03.2017',
                        'type': 'depot'}}
        self.assertEqual(self.wrapper._parse_overview(BeautifulSoup(html, 'html5lib')), e_result)

    def test_025_get_document_links(self, mock_browser):
        """ test DKBRobo._get_document_links() method """
        html = read_file(self.dir_path + '/mocks/doclinks.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        e_result = {'Kontoauszug Nr. 003_2017 zu Konto 87654321': 'https://www.ib.dkb.de/doc-2',
                    'Kontoauszug Nr. 003_2017 zu Konto 12345678': 'https://www.ib.dkb.de/doc-1'}
        self.assertEqual(self.wrapper._get_document_links('http://foo.bar/foo'), e_result)

    @patch('dkb_robo.legacy.Wrapper._update_downloadstate')
    @patch('dkb_robo.legacy.Wrapper._get_document')
    def test_026_get_document_links(self, mock_doc, mock_updow, mock_browser):
        """ test DKBRobo._get_document_links() method """
        html = read_file(self.dir_path + '/mocks/doclinks-2.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_doc.return_value = (200, 'fname', ['foo'])
        e_result = {'Kontoauszug Nr. 003_2017 zu Konto 12345678': {'rcode': 200, 'link': 'https://www.ib.dkb.de/doc-1', 'fname': 'fname'}, 'Kontoauszug Nr. 003_2017 zu Konto 87654321': {'rcode': 200, 'link': 'https://www.ib.dkb.de/doc-2', 'fname': 'fname'}}
        self.assertEqual(e_result, self.wrapper._get_document_links('http://foo.bar/foo', path='path'))
        self.assertTrue(mock_updow.called)

    @patch('dkb_robo.legacy.Wrapper._update_downloadstate')
    @patch('dkb_robo.legacy.Wrapper._get_document')
    def test_027_get_document_links(self, mock_doc, mock_updow, mock_browser):
        """ test DKBRobo._get_document_links() method """
        html1 = read_file(self.dir_path + '/mocks/doclinks-3.html')
        html2 = read_file(self.dir_path + '/mocks/doclinks-2.html')
        mock_browser.get_current_page.side_effect = [BeautifulSoup(html1, 'html5lib'), BeautifulSoup(html2, 'html5lib')]
        mock_browser.open.return_value = True
        mock_doc.return_value = (None, 'fname', ['foo'])
        e_result = {'Kontoauszug Nr. 003_2017 zu Konto 23456789': 'https://www.ib.dkb.de/doc-1',
                    'Kontoauszug Nr. 003_2017 zu Konto 12345678': 'https://www.ib.dkb.de/doc-1',
                    'Kontoauszug Nr. 003_2017 zu Konto 87654321': 'https://www.ib.dkb.de/doc-2',
                    'Kontoauszug Nr. 003_2017 zu Konto 98765432': 'https://www.ib.dkb.de/doc-2'}
        self.assertEqual(e_result, self.wrapper._get_document_links('http://foo.bar/foo', path='path'))
        self.assertFalse(mock_updow.called)

    @patch('dkb_robo.legacy.Wrapper._update_downloadstate')
    @patch('dkb_robo.legacy.Wrapper._get_document')
    def test_028_get_document_links(self, mock_doc, mock_updow, mock_browser):
        """ test DKBRobo._get_document_links() method no html return """
        mock_browser.get_current_page.return_value = None
        mock_browser.open.return_value = True
        mock_doc.return_value = (None, 'fname', ['foo'])
        e_result = {}
        self.assertEqual(e_result, self.wrapper._get_document_links('http://foo.bar/foo', path='path'))
        self.assertFalse(mock_updow.called)

    @patch('dkb_robo.legacy.Wrapper._update_downloadstate')
    @patch('dkb_robo.legacy.Wrapper._get_document')
    def test_029_get_document_links(self, mock_doc, mock_updow, mock_browser):
        """ test DKBRobo._get_document_links() method  wrong html return """
        html = '<html><body>fooo</body></html>'
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_browser.open.return_value = True
        mock_doc.return_value = (None, 'fname', ['foo'])
        e_result = {}
        self.assertEqual(e_result, self.wrapper._get_document_links('http://foo.bar/foo', path='path'))
        self.assertFalse(mock_updow.called)

    @patch('dkb_robo.legacy.Wrapper._get_document_links')
    def test_030_scan_postbox(self, mock_doclinks, mock_browser):
        """ test DKBRobo.scan_postbox() method """
        html = read_file(self.dir_path + '/mocks/postbox.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_doclinks.return_value = {}
        e_result = {'Kreditkartenabrechnungen': {'documents': {}, 'name': 'Kreditkartenabrechnungen', 'details': 'https://www.ib.dkb.de/banking/postfach/Kreditkartenabrechnungen'},
                    'Mitteilungen': {'documents': {}, 'name': 'Mitteilungen', 'details': 'https://www.ib.dkb.de/banking/postfach/Mitteilungen'},
                    'Vertragsinformationen': {'documents': {}, 'name': 'Vertragsinformationen', 'details': 'https://www.ib.dkb.de/banking/postfach/Vertragsinformationen'}}
        self.assertEqual(self.wrapper.scan_postbox(), e_result)

    @patch('dkb_robo.legacy.Wrapper._get_document_links')
    def test_031_scan_postbox(self, mock_doclinks, mock_browser):
        """ test DKBRobo.scan_postbox() method """
        html = read_file(self.dir_path + '/mocks/postbox.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_doclinks.return_value = {}
        e_result = {'Kreditkartenabrechnungen':
                        {'documents': {},
                         'name': 'Kreditkartenabrechnungen',
                         'details': 'https://www.ib.dkb.de/banking/postfach/Kreditkartenabrechnungen'},
                    'Mitteilungen':
                        {'documents': {},
                         'name': 'Mitteilungen',
                         'details': 'https://www.ib.dkb.de/banking/postfach/Mitteilungen'},
                    'Vertragsinformationen':
                        {'documents': {},
                         'name': 'Vertragsinformationen',
                         'details': 'https://www.ib.dkb.de/banking/postfach/Vertragsinformationen'}}
        self.assertEqual(self.wrapper.scan_postbox(path='path'), e_result)

    @patch('dkb_robo.legacy.Wrapper._get_document_links')
    def test_032_scan_postbox(self, mock_doclinks, mock_browser):
        """ test DKBRobo.scan_postbox() method """
        html = read_file(self.dir_path + '/mocks/postbox-2.html')
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        mock_doclinks.return_value = {}
        e_result = {'Kreditkartenabrechnungen':
                        {'documents': {},
                         'name': 'Kreditkartenabrechnungen',
                         'details': 'https://www.ib.dkb.de/banking/postfach/Kreditkartenabrechnungen'},
                    'Mitteilungen':
                        {'documents': {},
                         'name': 'Mitteilungen',
                         'details': 'https://www.ib.dkb.de/banking/postfach/Mitteilungen'},
                    'Vertragsinformationen':
                        {'documents': {},
                         'name': 'Vertragsinformationen',
                         'details': 'https://www.ib.dkb.de/banking/postfach/Vertragsinformationen'}}
        self.assertEqual(self.wrapper.scan_postbox(archive=True), e_result)

    def test_033_get_tr_invalid(self, _unused):
        """ test DKBRobo.get_transactions() method with an invalid account type"""
        self.assertEqual(self.wrapper.get_transactions('url', 'foo', '01.03.2017', '02.03.2017'), [])

    @patch('dkb_robo.legacy.Wrapper._get_creditcard_transactions')
    def test_034_get_tr_cc(self, mock_cc_tran, _unused):
        """ test DKBRobo.get_transactions() method with an credit-card account"""
        mock_cc_tran.return_value = ['credit_card']
        self.assertEqual(self.wrapper.get_transactions('url', 'creditcard', '01.03.2017', '02.03.2017'), ['credit_card'])
        self.assertTrue(mock_cc_tran.called)

    @patch('dkb_robo.legacy.Wrapper._get_account_transactions')
    def test_035_get_tr_ac(self, mock_ca_tran, _unused):
        """ test DKBRobo.get_transactions() method with an checking account"""
        mock_ca_tran.return_value = ['account']
        self.assertEqual(self.wrapper.get_transactions('url', 'account', '01.03.2017', '02.03.2017'), ['account'])
        self.assertTrue(mock_ca_tran.called)

    @patch('dkb_robo.legacy.Wrapper._get_depot_status')
    def test_036_get_tr_ac(self, mock_dep_tran, _unused):
        """ test DKBRobo.get_transactions() method for a deptot """
        mock_dep_tran.return_value = ['dep']
        self.assertEqual(self.wrapper.get_transactions('url', 'depot', '01.03.2017', '02.03.2017'), ['dep'])
        self.assertTrue(mock_dep_tran.called)

    def test_037_parse_account_tr(self, _mock_browser):
        """ test DKBRobo.get_account_transactions for one page only """
        csv = read_file(self.dir_path + '/mocks/test_parse_account_tr.csv')
        result = [
            {'amount': 100.0, 'bdate': '01.03.2017', 'customerreferenz': 'Kundenreferenz1', 'customerreference': 'Kundenreferenz1', 'date': '01.03.2017', 'mandatereference': 'Mandatsreferenz1', 'peer': 'Auftraggeber1', 'peeraccount': 'Kontonummer1', 'peerbic': 'BLZ1', 'peerid': 'GID1', 'postingtext': 'Buchungstext1', 'reasonforpayment': 'Verwendungszweck1', 'text': 'Buchungstext1 Auftraggeber1 Verwendungszweck1', 'vdate': '01.03.2017'},
            {'amount': -200.0, 'bdate': '02.03.2017', 'customerreferenz': 'Kundenreferenz2', 'customerreference': 'Kundenreferenz2', 'date': '02.03.2017', 'mandatereference': 'Mandatsreferenz2', 'peer': 'Auftraggeber2', 'peeraccount': 'Kontonummer2', 'peerbic': 'BLZ2', 'peerid': 'GID2', 'postingtext': 'Buchungstext2', 'reasonforpayment': 'Verwendungszweck2', 'text': 'Buchungstext2 Auftraggeber2 Verwendungszweck2', 'vdate': '02.03.2017'},
            {'amount': 3000.0, 'bdate': '03.03.2017', 'customerreferenz': 'Kundenreferenz3', 'customerreference': 'Kundenreferenz3', 'date': '03.03.2017', 'mandatereference': 'Mandatsreferenz3', 'peer': 'Auftraggeber3', 'peeraccount': 'Kontonummer3', 'peerbic': 'BLZ3', 'peerid': 'GID3', 'postingtext': 'Buchungstext3', 'reasonforpayment': 'Verwendungszweck3', 'text': 'Buchungstext3 Auftraggeber3 Verwendungszweck3', 'vdate': '03.03.2017'},
            {'amount': -4000.0, 'bdate': '04.03.2017', 'customerreferenz': 'Kundenreferenz4', 'customerreference': 'Kundenreferenz4', 'date': '04.03.2017', 'mandatereference': 'Mandatsreferenz4', 'peer': 'Auftraggeber4', 'peeraccount': 'Kontonummer4', 'peerbic': 'BLZ4', 'peerid': 'GID4', 'postingtext': 'Buchungstext4', 'reasonforpayment': 'Verwendungszweck4', 'text': 'Buchungstext4 Auftraggeber4 Verwendungszweck4', 'vdate': '04.03.2017'}]

        self.assertEqual(result, self.wrapper._parse_account_transactions(csv))

    def test_038_parse_no_account_tr(self, _mock_browser):
        """ test DKBRobo.get_account_transactions for one page only """
        csv = read_file(self.dir_path + '/mocks/test_parse_no_account_tr.csv')
        self.assertEqual(self.wrapper._parse_account_transactions(csv), [])

    def test_039_parse_dkb_cc_tr(self, _mock_browser):
        """ test DKBRobo._parse_cc_transactions """
        csv = read_file(self.dir_path + '/mocks/test_parse_dkb_cc_tr.csv')
        self.assertEqual(self.wrapper._parse_cc_transactions(csv), [{'amount': -100.00,
                                                                'amount_original': '-110',
                                                                'bdate': '01.03.2017',
                                                                'show_date': '01.03.2017',
                                                                'store_date': '01.03.2017',
                                                                'text': 'AAA',
                                                                'vdate': '01.03.2017'},
                                                               {'amount': -200.00,
                                                                'amount_original': '-210',
                                                                'bdate': '02.03.2017',
                                                                'show_date': '02.03.2017',
                                                                'store_date': '02.03.2017',
                                                                'text': 'BBB',
                                                                'vdate': '02.03.2017'},
                                                               {'amount': -300.00,
                                                                'amount_original': '-310',
                                                                'bdate': '03.03.2017',
                                                                'show_date': '03.03.2017',
                                                                'store_date': '03.03.2017',
                                                                'text': 'CCC',
                                                                'vdate': '03.03.2017'}])

    def test_040_parse_no_cc_tr(self, _mock_browser):
        """ test DKBRobo._parse_cc_transactions """
        csv = read_file(self.dir_path + '/mocks/test_parse_no_cc_tr.csv')
        self.assertEqual(self.wrapper._parse_cc_transactions(csv), [])

    def test_041_get_financial_statement(self, mock_browser):
        """ get financial statement """
        html = '<html><head>header</head><body>body</body></html>'
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        self.assertEqual('<html><head></head><body>headerbody</body></html>', str(self.wrapper._get_financial_statement()))

    def test_042_get_financial_statement(self, mock_browser):
        """ get financial statement with tan_insert """
        html = '<html><head>header</head><body>body</body></html>'
        self.wrapper.tan_insert = True
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        self.assertEqual('<html><head></head><body>headerbody</body></html>', str(self.wrapper._get_financial_statement()))

    @patch('dkb_robo.legacy.Wrapper._parse_account_transactions')
    def test_043_get_account_transactions(self, mock_parse, mock_browser):
        """ test _get_account_transactions """
        mock_browser.get_current_page.return_value = 'mock_browser'
        mock_parse.return_value = 'mock_parse'
        self.assertEqual('mock_parse', self.wrapper._get_account_transactions('url', 'date_from', 'date_to'))

    @patch('dkb_robo.legacy.Wrapper._parse_account_transactions')
    def test_044_get_account_transactions(self, mock_parse, mock_browser):
        """ test _get_account_transactions """
        mock_browser.get_current_page.return_value = 'mock_browser'
        mock_parse.return_value = 'mock_parse'
        self.assertEqual('mock_parse', self.wrapper._get_account_transactions('url', 'date_from', 'date_to', transaction_type='reserved'))

    @patch('dkb_robo.legacy.Wrapper._parse_cc_transactions')
    def test_045_get_cc_transactions(self, mock_parse, mock_browser):
        """ test _get_account_transactions """
        mock_browser.get_current_page.return_value = 'mock_browser'
        mock_parse.return_value = 'mock_parse'
        self.assertEqual('mock_parse', self.wrapper._get_creditcard_transactions('url', 'date_from', 'date_to'))

    @patch('dkb_robo.legacy.Wrapper._parse_cc_transactions')
    def test_046_get_cc_transactions(self, mock_parse, mock_browser):
        """ test _get_account_transactions """
        mock_browser.get_current_page.return_value = 'mock_browser'
        mock_parse.return_value = 'mock_parse'
        self.assertEqual('mock_parse', self.wrapper._get_creditcard_transactions('url', 'date_from', 'date_to', transaction_type='reserved'))

    def test_047_update_downloadstate(self, _unused):
        """ test update downloadstats """
        url = 'https://www.ib.dkb.de/DkbTransactionBanking/content/mailbox/MessageList/%24{1}.xhtml?$event=updateDownloadState&row=1'
        self.assertFalse(self.wrapper._update_downloadstate(link_name='link_name', url=url))

    def test_048_update_downloadstate(self, _unused):
        """ test update downloadstats """
        url = 'https://www.ib.dkb.de/DkbTransactionBanking/content/mailbox/MessageList/%24{1}.xhtml?$event=updateDownloadState&row=1'
        self.assertFalse(self.wrapper._update_downloadstate(link_name='Kontoauszüge', url=url))

    @patch('dkb_robo.legacy.generate_random_string')
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_049_get_document(self, mock_exists, mock_makedir, mock_rand, _unused):
        """ test get_document create path """
        mock_exists.return_value = False
        mock_rand.return_value = 'mock_rand'
        self.assertEqual((None, 'path/mock_rand.pdf', []), self.wrapper._get_document('folder_url', 'path', 'url', [], False))
        self.assertTrue(mock_makedir.called)
        self.assertTrue(mock_rand.called)

    @patch('dkb_robo.legacy.generate_random_string')
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_050_get_document(self, mock_exists, mock_makedir, mock_rand, _unused):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_rand.return_value = 'mock_rand'
        self.assertEqual((None, 'path/mock_rand.pdf', []), self.wrapper._get_document('folder_url', 'path', 'url', [], False))
        self.assertFalse(mock_makedir.called)
        self.assertTrue(mock_rand.called)

    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('re.findall')
    @patch('dkb_robo.utilities.generate_random_string')
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_051_get_document(self, mock_exists, mock_makedir, mock_rand, mock_re, mock_browser):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_rand.return_value = 'mock_rand'
        mock_browser.open.return_value.headers = {'Content-Disposition': ['foo', 'bar']}
        mock_browser.open.return_value.status_code = 200
        mock_re.return_value = ['mock_re.pdf', 'mock_re2.pdf']
        self.assertEqual((200, 'path/mock_re.pdf', ['mock_re.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', [], False))
        self.assertFalse(mock_makedir.called)

    @patch('dkb_robo.legacy.datetime', Mock(now=lambda: date(2022, 9, 30)))
    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('re.findall')
    @patch('dkb_robo.utilities.generate_random_string')
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_052_get_document(self, mock_exists, mock_makedir, mock_rand, mock_re, mock_browser):
        """ test get_document override """
        mock_exists.return_value = True
        mock_rand.return_value = 'mock_rand'
        mock_browser.open.return_value.headers = {'Content-Disposition': ['foo', 'bar']}
        mock_browser.open.return_value.status_code = 200
        mock_re.return_value = ['mock_re.pdf', 'mock_re2.pdf']
        self.assertEqual((200, 'path/2022-09-30-00-00-00_mock_re.pdf', ['mock_re.pdf', '2022-09-30-00-00-00_mock_re.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', ['mock_re.pdf'], False))
        self.assertFalse(mock_makedir.called)

    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_053_get_document(self, mock_exists, mock_makedir, mock_browser):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_browser.open.return_value.headers = {'Content-Disposition': 'inline; filename=Mitteilung_%c3%bcber_steigende_Sollzinss%c3%a4tze_ab_01.10.2022.pdf'}
        mock_browser.open.return_value.status_code = 200
        self.assertEqual((200, 'path/Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf', ['Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', [], False))
        self.assertFalse(mock_makedir.called)

    @patch('dkb_robo.legacy.datetime', Mock(now=lambda: date(2022, 9, 30)))
    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_054_get_document(self, mock_exists, mock_makedir, mock_browser):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_browser.open.return_value.headers = {'Content-Disposition': 'inline; filename=Mitteilung_%c3%bcber_steigende_Sollzinss%c3%a4tze_ab_01.10.2022.pdf'}
        mock_browser.open.return_value.status_code = 200
        self.assertEqual((200, 'path/2022-09-30-00-00-00_Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf', ['Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf', '2022-09-30-00-00-00_Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', ['Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf'], False))
        self.assertFalse(mock_makedir.called)

    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_055_get_document(self, mock_exists, mock_makedir, mock_browser):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_browser.open.return_value.headers = {'Content-Disposition': 'inline; filename=foo.pdf'}
        mock_browser.open.return_value.status_code = 200
        self.assertEqual((200, 'path/foo.pdf', ['foo.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', [], False))
        self.assertFalse(mock_makedir.called)

    @patch('dkb_robo.legacy.datetime', Mock(now=lambda: date(2022, 9, 30)))
    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_056_get_document(self, mock_exists, mock_makedir, mock_browser):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_browser.open.return_value.headers = {'Content-Disposition': 'inline; filename=foo.pdf'}
        mock_browser.open.return_value.status_code = 200
        self.assertEqual((200, 'path/2022-09-30-00-00-00_foo.pdf', ['foo.pdf', '2022-09-30-00-00-00_foo.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', ['foo.pdf'], False))
        self.assertFalse(mock_makedir.called)

    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('urllib.parse.unquote')
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_057_get_document(self, mock_exists, mock_makedir, mock_parse, mock_browser):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_browser.open.return_value.headers = {'Content-Disposition': 'inline; filename=Mitteilung_%c3%bcber_steigende_Sollzinss%c3%a4tze_ab_01.10.2022.pdf'}
        mock_browser.open.return_value.status_code = 200
        mock_parse.side_effect = [Exception('exc1')]
        self.assertEqual((200, 'path/Mitteilung_%c3%bcber_steigende_Sollzinss%c3%a4tze_ab_01.10.2022.pdf', ['Mitteilung_%c3%bcber_steigende_Sollzinss%c3%a4tze_ab_01.10.2022.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', [], False))
        self.assertFalse(mock_makedir.called)

    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_058_get_document(self, mock_exists, mock_makedir, mock_browser):
        """ test get_document prepend string """
        mock_exists.return_value = True
        mock_browser.open.return_value.headers = {'Content-Disposition': 'inline; filename=Mitteilung_%c3%bcber_steigende_Sollzinss%c3%a4tze_ab_01.10.2022.pdf'}
        mock_browser.open.return_value.status_code = 200
        self.assertEqual((200, 'path/prepend_Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf', ['prepend_Mitteilung_über_steigende_Sollzinssätze_ab_01.10.2022.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', [], 'prepend_'))
        self.assertFalse(mock_makedir.called)

    @patch("builtins.open", mock_open(read_data='test'), create=True)
    @patch('os.makedirs')
    @patch('os.path.exists')
    def test_059_get_document(self, mock_exists, mock_makedir, mock_browser):
        """ test get_document create path """
        mock_exists.return_value = True
        mock_browser.open.return_value.headers = {'Content-Disposition': 'inline; filename=foo.pdf'}
        mock_browser.open.return_value.status_code = 200
        self.assertEqual((200, 'path/prepend_foo.pdf', ['prepend_foo.pdf']), self.wrapper._get_document('folder_url', 'path', 'url', [], 'prepend_'))
        self.assertFalse(mock_makedir.called)

    @patch('builtins.input')
    def test_060_ctan_check(self, mock_input, mock_browser):
        """ test ctan_check """
        mock_input.return_value = 'tan'
        html = '<html><head>header</head><body><ol><li>li</li></ol></body></html>'
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        self.assertTrue(self.wrapper._ctan_check('soup'))

    @patch('builtins.input')
    def test_061_ctan_check(self, mock_input, mock_browser):
        """ test ctan_check """
        mock_input.return_value = 'tan'
        html = '<html><head>header</head><body>body</body></html>'
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        self.assertTrue(self.wrapper._ctan_check('soup'))

    @patch('builtins.input')
    def test_062_ctan_check(self, mock_input, mock_browser):
        """ test ctan_check wrong tan """
        mock_input.return_value = 'tan'
        html = '<html><head>header</head><body><div class="clearfix module text errorMessage">div</div></body></html>'
        mock_browser.get_current_page.return_value = BeautifulSoup(html, 'html5lib')
        with self.assertRaises(Exception) as err:
            self.wrapper._ctan_check('soup')
        self.assertEqual('Login failed due to wrong TAN', str(err.exception))

    @patch('dkb_robo.legacy.Wrapper._check_confirmation')
    def test_063_login_confirm(self, mock_confirm, mock_browser, ):
        """ test login confirmed check_cofirmation returns true """
        mock_browser.open.return_value.json.return_value = {"foo": "bar"}
        mock_confirm.return_value = True
        self.assertTrue(self.wrapper._login_confirm())

    @patch('time.sleep', return_value=None)
    @patch('dkb_robo.legacy.Wrapper._check_confirmation')
    def test_064_login_confirm(self, mock_confirm, _mock_sleep, mock_browser):
        """ test login confirmed check_cofirmation returns multiple false but then true """
        mock_browser.open.return_value.json.return_value = {"foo": "bar"}
        mock_confirm.side_effect = [False, False, False, True]
        self.assertTrue(self.wrapper._login_confirm())

    @patch('time.sleep', return_value=None)
    @patch('dkb_robo.legacy.Wrapper._check_confirmation')
    def test_065_login_confirm(self, mock_confirm, _mock_sleep, mock_browser):
        """ test login confirmed  """
        mock_browser.open.return_value.json.return_value = {"foo": "bar"}
        mock_confirm.return_value = False
        with self.assertRaises(Exception) as err:
            self.assertFalse(self.wrapper._login_confirm())
        self.assertEqual('No session confirmation after 120 polls', str(err.exception))

    def test_066_login_confirm(self, mock_browser):
        """ test login confirmed - exception when getting the token """
        mock_browser.open.return_value.json.return_value = {"foo": "bar"}
        mock_browser.get_current_page.side_effect = Exception('exc')
        with self.assertRaises(Exception) as err:
            self.assertTrue(self.wrapper._login_confirm())
        self.assertEqual('Error while getting the confirmation page', str(err.exception))

    def test_067_check_confirmation(self, _unused):
        """ test confirmation """
        result = {'foo': 'bar'}
        with self.assertRaises(Exception) as err:
            self.wrapper._check_confirmation(result, 1)
        self.assertEqual('Error during session confirmation', str(err.exception))

    def test_068_check_confirmation(self, _unused):
        """ test confirmation state expired"""
        result = {'state': 'EXPIRED'}
        with self.assertRaises(Exception) as err:
            self.wrapper._check_confirmation(result, 1)
        self.assertEqual('Session expired', str(err.exception))

    def test_069_check_confirmation(self, _unused):
        """ test confirmation state processed"""
        result = {'state': 'PROCESSED'}
        self.assertTrue(self.wrapper._check_confirmation(result, 1))

    def test_070_check_confirmation(self, _unused):
        """ test confirmation state unknown """
        result = {'state': 'UNK'}
        self.assertFalse(self.wrapper._check_confirmation(result, 1))

    def test_071_check_confirmation(self, _unused):
        """ test confirmation guiState expired"""
        result = {'guiState': 'EXPIRED'}
        with self.assertRaises(Exception) as err:
            self.wrapper._check_confirmation(result, 1)
        self.assertEqual('Session expired', str(err.exception))

    def test_072_check_confirmation(self, _unused):
        """ test confirmation guiState MAP_TO_EXIT"""
        result = {'guiState': 'MAP_TO_EXIT'}
        self.assertTrue(self.wrapper._check_confirmation(result, 1))

    def test_073_check_confirmation(self, _unused):
        """ test confirmation guiState unknown """
        result = {'guiState': 'UNK'}
        self.assertFalse(self.wrapper._check_confirmation(result, 1))

    def test_074_parse_depot_status_tr(self, _mock_browser):
        """ test DKBRobo._parse_cc_transactions """
        csv = read_file(self.dir_path + '/mocks/test_parse_depot.csv')
        result = [{'shares': 10.0, 'shares_unit': 'cnt1', 'isin_wkn': 'WKN1', 'text': 'Bezeichnung1', 'price': 11.0, 'win_loss': '', 'win_loss_currency': '', 'aquisition_cost': '', 'aquisition_cost_currency': '', 'dev_price': '', 'price_euro': 1110.1, 'availability': 'Frei'}, {'shares': 20.0, 'shares_unit': 'cnt2', 'isin_wkn': 'WKN2', 'text': 'Bezeichnung2', 'price': 12.0, 'win_loss': '', 'win_loss_currency': '', 'aquisition_cost': '', 'aquisition_cost_currency': '', 'dev_price': '', 'price_euro': 2220.2, 'availability': 'Frei'}]
        self.assertEqual(result, self.wrapper._parse_depot_status(csv))

    @patch('dkb_robo.legacy.Wrapper._parse_depot_status')
    def test_075_get_depot_status(self, mock_pds, _unused):
        """ test get depot status """
        mock_pds.return_value = 'mock_pds'
        self.assertEqual('mock_pds', self.wrapper._get_depot_status('url', 'fdate', 'tdate', 'booked'))

    @patch('dkb_robo.legacy.Wrapper._get_document')
    def test_076_legacy_download_document(self, mock_get_doc, _ununsed):
        """ test download document """
        html = read_file(self.dir_path + '/mocks/document_list.html')
        table = BeautifulSoup(html, 'html5lib')
        mock_get_doc.side_effect = my_side_effect
        class_filter = {}
        doc_dic = {'Name 04.01.2022': {'rcode': 200, 'link': 'https://www.ib.dkb.dehttps://www.dkb.de/DkbTransactionBanking/content/mailbox/MessageList.xhtml?$event=getMailboxAttachment&filename=Name+04.01.2022&row=0', 'fname': 'path/link_name'}}
        result = (doc_dic, [''])
        self.assertEqual(result, self.wrapper._download_document('folder_url', 'path', class_filter, 'link_name', table, False))

    @patch('dkb_robo.legacy.Wrapper._get_document')
    def test_077_legacy_download_document(self, mock_get_doc, _ununsed):
        """ test download document prepend date """
        html = read_file(self.dir_path + '/mocks/document_list.html')
        table = BeautifulSoup(html, 'html5lib')
        mock_get_doc.side_effect = my_side_effect
        class_filter = {}
        doc_dic = {'Name 04.01.2022': {'rcode': 200, 'link': 'https://www.ib.dkb.dehttps://www.dkb.de/DkbTransactionBanking/content/mailbox/MessageList.xhtml?$event=getMailboxAttachment&filename=Name+04.01.2022&row=0', 'fname': 'path/link_name'}}
        result = (doc_dic, ['2022-01-04_'])
        self.assertEqual(result, self.wrapper._download_document('folder_url', 'path', class_filter, 'link_name', table, True))

    @patch('dkb_robo.legacy.Wrapper._get_document')
    def test_078_legacy_download_document(self, mock_get_doc, _ununsed):
        """ test download document prepend date """
        html = read_file(self.dir_path + '/mocks/document_list-2.html')
        table = BeautifulSoup(html, 'html5lib')
        mock_get_doc.side_effect = my_side_effect
        class_filter = {}
        doc_dic = {'Name 04.01.2022': {'rcode': 200, 'link': 'https://www.ib.dkb.dehttps://www.dkb.de/DkbTransactionBanking/content/mailbox/MessageList.xhtml?$event=getMailboxAttachment&filename=Name+04.01.2022&row=0', 'fname': 'path/link_name'}}
        result = (doc_dic, [''])
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(result, self.wrapper._download_document('folder_url', 'path', class_filter, 'link_name', table, True))
        self.assertIn("ERROR:dkb_robo.legacy:Can't parse date, this could i.e. be for archived documents.", lcm.output)

    def test_079__get_formatted_date(self, _ununsed):
        """ test _get_formatted_date() prepend True """
        html = '<table><tr><td>foo</td><td class="abaxx-aspect-messageWithState-mailboxMessage-created">04.01.2022</td><td>bar</td></tr><table>'
        table = BeautifulSoup(html, 'html5lib')
        self.assertEqual('2022-01-04_', self.wrapper._get_formatted_date(True, table))

    def test_080__get_formatted_date(self, _ununsed):
        """ test _get_formatted_date() prepend False """
        html = '<table><tr><td>foo</td><td class="abaxx-aspect-messageWithState-mailboxMessage-created">04.01.2022</td><td>bar</td></tr><table>'
        table = BeautifulSoup(html, 'html5lib')
        self.assertEqual('', self.wrapper._get_formatted_date(False, table))

    def test_081__get_formatted_date(self, _ununsed):
        """ test _get_formatted_date() prepend False """
        html = '<table><tr><td>foo</td><td class="abaxx-aspect-messageWithState-mailboxMessage-created">fii</td><td>bar</td></tr><table>'
        table = BeautifulSoup(html, 'html5lib')
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual('', self.wrapper._get_formatted_date(True, table))
        self.assertIn("ERROR:dkb_robo.legacy:Can't parse date, this could i.e. be for archived documents.", lcm.output)

    def test_082_logout(self, _unused):
        """" test logout """
        self.wrapper.dkb_br = Mock()
        self.wrapper.dkb_br.open = Mock()
        self.assertFalse(self.wrapper.logout())
        self.assertTrue(self.wrapper.dkb_br.open.called)

    def test_083_logout(self, _unused):
        """" test logout """
        self.wrapper.dkb_br = None
        self.assertFalse(self.wrapper.logout())


if __name__ == '__main__':

    unittest.main()



# pylint: disable=r0913
""" legacy api """
# -*- coding: utf-8 -*-
import os
from datetime import datetime, timezone
import time
import re
import logging
from http import cookiejar
import csv
from typing import Dict, List, Tuple
from urllib import parse
import mechanicalsoup
import bs4
from dkb_robo.utilities import string2float, generate_random_string


logger = logging.getLogger(__name__)


class DKBRoboError(Exception):
    """ dkb-robo exception class """


class Wrapper(object):
    """ this is the wrapper for the legacy api """
    base_url = 'https://www.ib.dkb.de'
    dkb_user = None
    dkb_password = None
    dkb_br = None
    tan_insert = False
    proxies = {}

    def __init__(self, dkb_user: str = None, dkb_password: str = None, tan_insert: bool = False, proxies: Dict[str, str] = None):
        self.dkb_user = dkb_user
        self.dkb_password = dkb_password
        self.tan_insert = tan_insert
        self.proxies = proxies

    def _get_amount(self, cols: bs4.element.ResultSet, ontop: int) -> float:
        """ get link for transactions """
        logger.debug('Wrapper._get_amount()')
        amount = cols[3 + ontop].text.strip().replace('.', '')
        try:
            result = float(amount.replace(',', '.'))
        except Exception as _err:
            logger.error('Wrapper._parse_overview() convert amount: %s\n', _err)
            result = None

        logger.debug('Wrapper._get_amount() ended')
        return result

    def _check_confirmation(self, result: Dict[str, str], poll_id: int) -> bool:
        """ check if login has been confirmed via app """
        logger.debug('Wrapper._check_confirmation()\n')

        login_confirmed = False
        if 'state' in result:
            # new dkb mfa app
            logger.debug('mfa poll(id: %s status: %s)\n', poll_id, result['state'])
            # pylint: disable=R1723
            if result['state'] == 'PROCESSED':
                logger.debug('Session got confirmed...\n')
                login_confirmed = True
            elif result['state'] == 'EXPIRED':
                raise DKBRoboError('Session expired')
        elif 'guiState' in result:
            # legacy dkb app
            logger.debug('legacy poll(id: %s status: %s)\n', poll_id, result['guiState'])
            # pylint: disable=R1723
            if result['guiState'] == 'MAP_TO_EXIT':
                logger.debug('Session got confirmed...\n')
                login_confirmed = True
            elif result['guiState'] == 'EXPIRED':
                raise DKBRoboError('Session expired')
        else:
            raise DKBRoboError('Error during session confirmation')

        logger.debug('Wrapper._check_confirmation() ended with %s\n', login_confirmed)
        return login_confirmed

    def _ctan_check(self, _soup: str) -> bool:
        """ input of chiptan during login """
        logger.debug('Wrapper._ctan_check()\n')

        event_field = '$event'

        try:
            self.dkb_br.select_form('form[name="confirmForm"]')
            self.dkb_br[event_field] = 'tanVerification'
        except Exception as _err:  # pragma: no cover
            logger.debug('confirmForm not found\n')

        # open page to insert tan
        self.dkb_br.submit_selected()
        soup = self.dkb_br.get_current_page()

        login_confirm = False

        # select tan form
        self.dkb_br.select_form('#next')
        # print steps to be done
        olist = soup.find("ol")
        if olist:
            for li_ in olist.find_all('li'):
                print(li_.text.strip())
        else:
            print('Please open the TAN2GO app to get a TAN to be inserted below.')

        # ask for TAN
        self.dkb_br["tan"] = input("TAN: ")
        self.dkb_br[event_field] = 'next'

        # submit form and check response
        self.dkb_br.submit_selected()
        soup = self.dkb_br.get_current_page()

        # catch tan error
        # pylint: disable=R1720
        if soup.find("div", attrs={'class': 'clearfix module text errorMessage'}):
            raise DKBRoboError('Login failed due to wrong TAN')
        else:
            logger.debug('TAN is correct...\n')
            login_confirm = True

        logger.debug('Wrapper._ctan_check() ended with :%s\n', login_confirm)
        return login_confirm

    def _download_document(self, folder_url: str, path: str, class_filter: Dict[str, str], folder: str, table: 'bs4.BeautifulSoup', prepend_date: bool) -> Tuple[Dict, List]:
        """ document download """
        logger.debug('Wrapper._download_document()\n')
        document_dic = {}
        document_name_list = []

        tbody = table.find('tbody')
        for row in tbody.find_all('tr', class_filter):
            link = row.find('a')

            # get formatted date
            formatted_date = self._get_formatted_date(prepend_date, row)

            # download file
            if path:
                folder_path = f'{path}/{folder}'
                rcode, fname, document_name_list = self._get_document(folder_url, folder_path, self.base_url + link['href'], document_name_list, formatted_date)
                if rcode == 200:
                    # mark url as read
                    self._update_downloadstate(folder, self.base_url + link['href'])
                if rcode:
                    document_dic[link.contents[0]] = {'rcode': rcode, 'link': self.base_url + link['href'], 'fname': fname}
                else:
                    document_dic[link.contents[0]] = self.base_url + link['href']
            else:
                document_dic[link.contents[0]] = self.base_url + link['href']

        logger.debug('Wrapper._download_document()\n')
        return (document_dic, document_name_list)

    def _get_account_transactions(self, transaction_url: str, date_from: str, date_to: str, transaction_type: str = "booked") -> Dict:
        """ get transactions from an regular account for a certain amount of time """
        logger.debug('Wrapper._get_account_transactions(%s: %s/%s, %s)\n', transaction_url, date_from, date_to, transaction_type)

        self.dkb_br.open(transaction_url)
        self.dkb_br.select_form('#form1615473160_1')
        if transaction_type == 'reserved':
            self.dkb_br["slTransactionStatus"] = 1
        else:
            self.dkb_br["slTransactionStatus"] = 0
        self.dkb_br["searchPeriodRadio"] = 1
        self.dkb_br["slSearchPeriod"] = 1
        self.dkb_br["transactionDate"] = str(date_from)
        self.dkb_br["toTransactionDate"] = str(date_to)
        self.dkb_br.submit_selected()
        self.dkb_br.get_current_page()
        response = self.dkb_br.follow_link('csvExport')

        logger.debug('Wrapper._get_account_transactions() ended\n')
        return self._parse_account_transactions(response.content)

    def _get_cc_limits(self, form: bs4.element.Tag) -> Dict[str, str]:
        """ get credit card limits """
        logger.debug('Wrapper._get_cc_limits()\n')

        limit_dic = {}
        table = form.find('table', attrs={'class': 'multiColumn'})
        if table:
            rows = table.find_all("tr")
            for row in rows:
                cols = row.find_all("td")
                tmp = row.find("th")
                if cols:
                    try:
                        limit = tmp.find('span').text.strip()
                        account = cols[0].find('div', attrs={'class': 'minorLine'}).text.strip()
                        limit_dic[account] = string2float(limit)
                    except Exception as _err:
                        logger.error('Wrapper.get_credit_limits() get credit card limits: %s\n', _err)

        logger.debug('Wrapper._get_cc_limits() ended\n')
        return limit_dic

    def _get_checking_account_limit(self, form: bs4.element.Tag) -> Dict[str, str]:
        """ get checking account limits """
        logger.debug('Wrapper._get_checking_account_limit()\n')

        limit_dic = {}
        table = form.find('table', attrs={'class': 'dropdownAnchor'})
        if table:
            for row in table.find_all("tr"):
                cols = row.find_all("td")
                tmp = row.find("th")
                if cols:
                    limit = tmp.find('span').text.strip()
                    account = cols[0].find('div', attrs={'class': 'minorLine'}).text.strip()
                    limit_dic[account] = string2float(limit)

        logger.debug('Wrapper._get_checking_account_limit() ended\n')
        return limit_dic

    def _get_creditcard_transactions(self, transaction_url: str, date_from: str, date_to: str, transaction_type: str = "booked") -> Dict[str, str]:
        """ get transactions from an regular account for a certain amount of time """
        logger.debug('Wrapper._legacy_get_creditcard_transactions(%s: %s/%s, %s)\n', transaction_url, date_from, date_to, transaction_type)
        # get credit card transaction form yesterday
        self.dkb_br.open(transaction_url)
        self.dkb_br.select_form('#form1579108072_1')
        if transaction_type == 'reserved':
            self.dkb_br["slTransactionStatus"] = 1
        else:
            self.dkb_br["slTransactionStatus"] = 0
        self.dkb_br["slSearchPeriod"] = 0
        self.dkb_br["filterType"] = 'DATE_RANGE'
        self.dkb_br["postingDate"] = str(date_from)
        self.dkb_br["toPostingDate"] = str(date_to)
        self.dkb_br.submit_selected()

        response = self.dkb_br.follow_link('csvExport')

        logger.debug('Wrapper._get_creditcard_transactions() ended')
        return self._parse_cc_transactions(response.content)

    def _get_depot_status(self, transaction_url: str, date_from: str, date_to: str, transaction_type: str = "booked") -> Dict[str, str]:
        """ get depoot status """
        logger.debug('Wrapper..get_depot_transactions(%s: %s/%s, %s)\n', transaction_url, date_from, date_to, transaction_type)

        self.dkb_br.open(transaction_url)
        response = self.dkb_br.follow_link('csvExport')

        logger.debug('Wrapper._get_creditcard_transactions() ended´\n')
        return self._parse_depot_status(response.content)

    def _get_document(self, folder_url: str, path: str, url: str, document_name_list: List[str], formatted_date: str) -> Tuple[int, str, List[str]]:
        """ get download document from postbox """
        logger.debug('Wrapper._get_document(%s)\n', url)

        # create directory if not existing
        if not os.path.exists(path):
            logger.info('Wrapper._get_document() Create directory %s\n', path)
            os.makedirs(path)

        # fetch file
        response = self.dkb_br.open(folder_url)
        response = self.dkb_br.open(url)

        # gt filename from response header
        fname = ''
        if "Content-Disposition" in response.headers.keys():
            logger.debug('Wrapper._get_document(): response.header: %s\n', response.headers)
            # unquote filename to cover german umlaut including a fallback
            try:
                fname = parse.unquote(re.findall("filename=(.+)", response.headers["Content-Disposition"])[0])
            except Exception as _err:
                logger.debug('Wrapper._get_document(): error during filename conversion: %s\n', _err)
                fname = re.findall("filename=(.+)", response.headers["Content-Disposition"])[0]

            if fname in document_name_list:
                # rename to avoid overrides
                logger.debug('Wrapper._get_document(): adding datetime to avoid overrides.\n')
                now = datetime.now()
                fname = f'{now.strftime("%Y-%m-%d-%H-%M-%S")}_{fname}'

            if formatted_date:
                fname = f'{formatted_date}{fname}'

            # log filename
            logger.debug('Wrapper._get_document(): filename: %s\n', fname)

            # dump content to file
            logger.debug('Wrapper._get_document() content-length: %s\n', len(response.content))
            with open(f'{path}/{fname}', 'wb') as pdf_file:
                pdf_file.write(response.content)
            result = response.status_code
            document_name_list.append(fname)
        else:
            fname = f'{generate_random_string(20)}.pdf'
            result = None

        logger.debug('Wrapper._get_document() ended')
        return result, f'{path}/{fname}', document_name_list

    def _get_document_links(self, url: str, path: str = None, link_name: str = None, select_all: bool = False, prepend_date: bool = False) -> Dict[str, str]:
        """ create a dictionary of the documents stored in a pbost folder """
        # pylint: disable=R0914
        logger.debug('Wrapper._get_document_links(%s)\n', url)
        document_dic = {}

        # set download filter if there is a need to do so
        if path and not select_all:
            class_filter = {'class': 'mbo-messageState-unread'}
        else:
            class_filter = {}

        self.dkb_br.open(url)
        # create a list of documents to avoid overrides
        document_name_list = []

        next_url = url

        while True:
            soup = self.dkb_br.get_current_page()
            if soup:
                table = soup.find('table', attrs={'class': 'widget widget abaxx-table expandableTable expandableTable-with-sort'})
                if table:
                    (tmp_dic, tmp_list, ) = self._download_document(next_url, path, class_filter, link_name, table, prepend_date)
                    document_name_list.extend(tmp_list)
                    document_dic.update(tmp_dic)

                next_site = soup.find('span', attrs={'class': 'pager-navigator-next'})
                if next_site:
                    next_url = self.base_url + next_site.find('a')['href']
                    self.dkb_br.open(next_url)
                else:
                    break  # pragma: no cover
            else:
                break  # pragma: no cover

        logger.debug('Wrapper._get_document_links()')
        return document_dic

    def _get_evtdetails_link(self, cols: bs4.element.ResultSet, ontop: int) -> str:
        """ get link for details """
        logger.debug('Wrapper.get_evt_details()')

        try:
            link = cols[4 + ontop].find('a', attrs={'class': 'evt-details'})
            details_link = self.base_url + link['href']
        except Exception as _err:
            logger.error('Wrapper._parse_overview() get link: %s\n', _err)
            details_link = None

        logger.debug('Wrapper.get_evt_details() ended')
        return details_link

    def _get_financial_statement(self) -> bs4.BeautifulSoup:
        """ get finanical statement """
        logger.debug('Wrapper._get_financial_statement()\n')

        statement_url = self.base_url + '/ssohl/DkbTransactionBanking/content/banking/financialstatus/FinancialComposite/FinancialStatus.xhtml?$event=init'

        self.dkb_br.open(statement_url)
        soup = self.dkb_br.get_current_page()

        logger.debug('Wrapper._get_financial_statement() ended\n')
        return soup

    def _get_formatted_date(self, prepend_date: bool, row: bs4.element.Tag) -> str:
        """ get document date for prepending """
        logger.debug('Wrapper._get_formatted_date()\n')

        formatted_date = ""
        if prepend_date:
            try:
                creation_date = row.find('td', attrs={'class': 'abaxx-aspect-messageWithState-mailboxMessage-created'}).text
                creation_date_components = creation_date.split(".")
                formatted_date = f'{creation_date_components[2]}-{creation_date_components[1]}-{creation_date_components[0]}_'
            except Exception:
                logger.error("Can't parse date, this could i.e. be for archived documents.")

        logger.debug('Wrapper._get_formatted_date() ended\n')
        return formatted_date

    def _get_transaction_link(self, cols: bs4.element.ResultSet, ontop: int, account_number: str) -> Tuple[str, str]:
        """ get link for transactions """
        logger.debug('Wrapper.get_transaction_link()')

        account_type = None
        transaction_link = None

        link = cols[4 + ontop].find('a', attrs={'class': 'evt-paymentTransaction'})
        if link:
            # thats a cash account or a credit card
            if 'cash' in cols[4 + ontop].text.strip().lower() or account_number.startswith('DE'):
                # this is a cash account
                account_type = 'account'
            else:
                # this is a credit card
                account_type = 'creditcard'
            transaction_link = self.base_url + link['href']
        else:
            try:
                # thats a depot
                account_type = 'depot'
                link = cols[4 + ontop].find('a', attrs={'class': 'evt-depot'})
                transaction_link = self.base_url + link['href']
            except Exception as _err:
                logger.error('Wrapper._parse_overview() parse depot: %s\n', _err)

        logger.debug('Wrapper._get_transaction_link() ended')
        return account_type, transaction_link

    def _login_confirm(self) -> bool:
        """ confirm login to dkb via app """
        logger.debug('Wrapper._login_confirm()\n')

        print('check your banking app and confirm login...')

        try:
            # get xsrf token
            soup = self.dkb_br.get_current_page()
            xsrf_token = soup.find('input', attrs={'name': 'XSRFPreventionToken'}).get('value')
        except Exception:
            # fallback
            soup = None

        # not confirmed by default
        login_confirmed = False

        if soup:
            # poll url
            poll_id = int(datetime.now(timezone.utc).timestamp() * 1e3)
            poll_url = self.base_url + soup.find("form", attrs={'id': 'confirmForm'}).get('action')
            for _cnt in range(120):
                # add id to pollurl
                poll_id += 1
                url = poll_url + '?$event=pollingVerification&$ignore.request=true&_=' + str(poll_id)
                result = self.dkb_br.open(url).json()
                login_confirmed = self._check_confirmation(result, poll_id)
                if login_confirmed:
                    break
                time.sleep(1.5)
            else:
                raise DKBRoboError("No session confirmation after 120 polls")

            post_data = {'$event': 'next', 'XSRFPreventionToken': xsrf_token}
            self.dkb_br.post(url=poll_url, data=post_data)
        else:
            raise DKBRoboError("Error while getting the confirmation page")

        logger.debug('Wrapper._login_confirm() ended\n')
        return login_confirmed

    def _new_instance(self, clientcookies=None) -> mechanicalsoup.stateful_browser.StatefulBrowser:
        """ creates a new browser instance """
        logger.debug('Wrapper._new_instance()\n')

        # create browser and cookiestore objects
        self.dkb_br = mechanicalsoup.StatefulBrowser()

        # set proxies
        if self.proxies:
            self.dkb_br.session.proxies = self.proxies
            self.dkb_br.session.verify = False

        dkb_cj = cookiejar.LWPCookieJar()
        self.dkb_br.set_cookiejar = dkb_cj

        # configure browser
        self.dkb_br.set_handle_equiv = True
        self.dkb_br.set_handle_redirect = True
        self.dkb_br.set_handle_referer = True
        self.dkb_br.set_handle_robots = False
        self.dkb_br.addheaders = [('User-agent', 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.2.8) Gecko/20100722 Firefox/3.6.8 GTB7.1 (.NET CLR 3.5.30729)'), ('Accept-Language', 'en-US,en;q=0.5'), ('Connection', 'keep-alive')]

        # initialize some cookies to fool dkb
        dkb_ck = cookiejar.Cookie(version=0, name='javascript', value='enabled', port=None, port_specified=False, domain=self.base_url, domain_specified=False, domain_initial_dot=False, path='/', path_specified=True, secure=False, expires=None, discard=True, comment=None, comment_url=None, rest={'HttpOnly': None}, rfc2109=False)
        dkb_cj.set_cookie(dkb_ck)
        dkb_ck = cookiejar.Cookie(version=0, name='BRSINFO_browserPlugins', value='NPSWF32_25_0_0_127.dll%3B', port=None, port_specified=False, domain=self.base_url, domain_specified=False, domain_initial_dot=False, path='/', path_specified=True, secure=False, expires=None, discard=True, comment=None, comment_url=None, rest={'HttpOnly': None}, rfc2109=False)
        dkb_cj.set_cookie(dkb_ck)
        dkb_ck = cookiejar.Cookie(version=0, name='BRSINFO_screen', value='width%3D1600%3Bheight%3D900%3BcolorDepth%3D24', port=None, port_specified=False, domain=self.base_url, domain_specified=False, domain_initial_dot=False, path='/', path_specified=True, secure=False, expires=None, discard=True, comment=None, comment_url=None, rest={'HttpOnly': None}, rfc2109=False)
        dkb_cj.set_cookie(dkb_ck)

        if clientcookies:
            logger.debug('_new_instance(): adding clientcookies.')
            for cookie in clientcookies:
                self.dkb_br.session.cookies.set_cookie(cookie)

        logger.debug('Wrapper._new_instance() ended\n')
        return self.dkb_br

    def _parse_account_transactions(self, transactions: bytes) -> List[str]:
        """ parses html code and creates a list of transactions included """
        logger.debug('Wrapper._parse_account_transactions()\n')

        # create empty list
        transaction_list = []
        # parse CSV
        for row in csv.reader(transactions.decode('latin-1').splitlines(), delimiter=';'):
            if len(row) == 12:
                # skip first line
                if row[0] != 'Buchungstag':
                    tmp_dic = {}

                    # data from CSV
                    tmp_dic['bdate'] = row[0]
                    tmp_dic['vdate'] = row[1]
                    # remove double spaces
                    tmp_dic['postingtext'] = ' '.join(row[2].split())
                    tmp_dic['peer'] = ' '.join(row[3].split())
                    tmp_dic['reasonforpayment'] = ' '.join(row[4].split())
                    tmp_dic['mandatereference'] = ' '.join(row[9].split())
                    tmp_dic['customerreferenz'] = ' '.join(row[10].split())
                    tmp_dic['customerreference'] = ' '.join(row[10].split())

                    tmp_dic['peeraccount'] = row[5]
                    tmp_dic['peerbic'] = row[6]
                    tmp_dic['peerid'] = row[8]

                    # reformat amount
                    tmp_dic['amount'] = string2float(row[7])

                    #  date is only for backwards compatibility
                    tmp_dic['date'] = row[0]
                    tmp_dic['text'] = f"{tmp_dic['postingtext']} {tmp_dic['peer']} {tmp_dic['reasonforpayment']}"

                    # append dic to list
                    transaction_list.append(tmp_dic)

        logger.debug('Wrapper._parse_account_transactions() ended\n')
        return transaction_list

    def _parse_cc_transactions(self, transactions: bytes) -> List[str]:
        """ parses html code and creates a list of transactions included """
        logger.debug('Wrapper._parse_cc_transactions()\n')

        # create empty list
        transaction_list = []

        # parse CSV
        for row in csv.reader(transactions.decode('latin-1').splitlines(), delimiter=';'):
            if len(row) == 7:
                # skip first line
                if row[1] != 'Wertstellung':
                    tmp_dic = {}
                    tmp_dic['vdate'] = row[1]
                    tmp_dic['show_date'] = row[1]
                    tmp_dic['bdate'] = row[2]
                    tmp_dic['store_date'] = row[2]
                    tmp_dic['text'] = row[3]
                    tmp_dic['amount'] = string2float(row[4])
                    tmp_dic['amount_original'] = row[5].replace('.', '').replace(',', '.')
                    # append dic to list
                    transaction_list.append(tmp_dic)

        logger.debug('Wrapper._parse_cc_transactions() ended\n')
        return transaction_list

    def _parse_depot_status(self, transactions: bytes) -> List[str]:
        """ parses html code and creates a list of stocks included  """
        logger.debug('Wrapper._parse_depot_status()\n')

        # create empty list
        stocks_list = []

        # parse CSV
        entry_row_length = 13
        for row in csv.reader(transactions.decode('latin-1').splitlines(), delimiter=';'):
            if len(row) == entry_row_length:
                # skip header line
                if row[0] != 'Bestand':
                    tmp_dic = {}
                    tmp_dic['shares'] = string2float(row[0])
                    tmp_dic['shares_unit'] = row[1]
                    tmp_dic['isin_wkn'] = row[2]
                    tmp_dic['text'] = row[3]
                    tmp_dic['price'] = string2float(row[4])
                    tmp_dic['win_loss'] = row[5]
                    tmp_dic['win_loss_currency'] = row[6]
                    tmp_dic['aquisition_cost'] = row[7]
                    tmp_dic['aquisition_cost_currency'] = row[8]
                    tmp_dic['dev_price'] = row[9]
                    tmp_dic['price_euro'] = string2float(row[10])
                    tmp_dic['availability'] = row[11]

                    # append dic to list
                    stocks_list.append(tmp_dic)

        logger.debug('Wrapper._parse_depot_status() ended\n')
        return stocks_list

    def _parse_overview(self, soup: bs4.BeautifulSoup) -> Dict[str, str]:
        """ creates a dictionary including account information """
        logger.debug('Wrapper._parse_overview()\n')

        overview_dic = {}
        counter = 0
        ontop = 0
        for row in soup.find_all("tr", attrs={'class': 'mainRow'}):
            overview_dic[counter] = {}
            cols = row.find_all("td")

            # check if we have accounts from other banks in overview
            # in this case we need to shift columns by one
            if cols[0].find("img"):
                ontop = 1

            # account name
            overview_dic[counter]['name'] = cols[0 + ontop].find('div').text.strip()

            # account number
            overview_dic[counter]['account'] = cols[1 + ontop].text.strip()
            # date
            overview_dic[counter]['date'] = cols[2 + ontop].text.strip()

            # amount
            amount = self._get_amount(cols, ontop)
            if amount:
                overview_dic[counter]['amount'] = amount

            # get link for transactions
            (account_type, transaction_link) = self._get_transaction_link(cols, ontop, overview_dic[counter]['account'])
            if account_type:
                overview_dic[counter]['type'] = account_type
            if transaction_link:
                overview_dic[counter]['transactions'] = transaction_link

            # get link for details
            details_link = self._get_evtdetails_link(cols, ontop)
            if details_link:
                overview_dic[counter]['details'] = details_link

            # increase counter
            counter += 1

        logger.debug('Wrapper._parse_overview() ended\n')
        return overview_dic

    def _update_downloadstate(self, link_name: str, url: str) -> Dict[str, str]:
        """ mark document as read """
        logger.debug('Wrapper._update_downloadstate(%s, %s)\n', link_name, url)

        # get row number to be marked as read
        row_num = parse.parse_qs(parse.urlparse(url).query)['row'][0]
        # construct url
        if link_name == 'Kontoauszüge':
            mark_link = 'kontoauszuege'
        else:
            mark_link = link_name.lower()
        mark_url = f'{self.base_url}/DkbTransactionBanking/content/mailbox/MessageList/%24{mark_link}.xhtml?$event=updateDownloadState&row={row_num}'
        # mark document by fetch url
        _response = self.dkb_br.open(mark_url)  # lgtm [py/unused-local-variable]
        logger.debug('Wrapper._update_downloadstate() ended\n')

    def get_credit_limits(self) -> Dict[str, str]:
        """ create a dictionary of credit limits of the different accounts """
        logger.debug('Wrapper.get_credit_limits()\n')

        limit_url = self.base_url + '/ssohl/DkbTransactionBanking/content/service/CreditcardLimit.xhtml'
        self.dkb_br.open(limit_url)

        soup = self.dkb_br.get_current_page()
        form = soup.find('form', attrs={'id': 'form597962073_1'})

        if form:
            # get checking account limits
            limit_dic = self._get_checking_account_limit(form)

            # get credit cards limits
            limit_dic.update(self._get_cc_limits(form))
        else:
            limit_dic = {}

        logger.debug('Wrapper.get_credit_limits() ended\n')
        return limit_dic

    def get_exemption_order(self) -> Dict[str, str]:
        """ returns a dictionary of the stored exemption orders """
        logger.debug('Wrapper.get_exemption_order()\n')

        exo_url = self.base_url + '/ssohl/DkbTransactionBanking/content/personaldata/ExemptionOrderOverview.xhtml'
        self.dkb_br.open(exo_url)

        soup = self.dkb_br.get_current_page()

        for lbr in soup.find_all("br"):
            lbr.replace_with("")
            # br.replace('<br />', ' ')

        table = soup.find('table', attrs={'class': 'expandableTable'})

        exo_dic = {}
        if table:
            count = 0
            for row in table.find_all("tr"):
                cols = row.find_all("td")
                if cols:
                    try:
                        count += 1
                        exo_dic[count] = {}
                        # description
                        description = re.sub(' +', ' ', cols[1].text.strip())
                        description = description.replace('\n', '')
                        description = description.replace('\r', '')
                        description = description.replace('  ', ' ')
                        exo_dic[count]['description'] = description

                        # validity
                        validity = re.sub(' +', ' ', cols[2].text.strip())
                        validity = validity.replace('\n', '')
                        validity = validity.replace('\r', '')
                        validity = validity.replace('  ', ' ')
                        exo_dic[count]['validity'] = validity
                        exo_dic[count]['amount'] = string2float(cols[3].text.strip().replace('EUR', ''))
                        exo_dic[count]['used'] = string2float(cols[4].text.strip().replace('EUR', ''))
                        exo_dic[count]['available'] = string2float(cols[5].text.strip().replace('EUR', ''))
                    except Exception as _err:
                        logger.error('Wrapper.get_exemption_order(): %s\n', _err)

        logger.debug('Wrapper.get_exemption_order() ended\n')
        return exo_dic

    def get_points(self) -> Dict[str, str]:
        """ returns the DKB points """
        logger.debug('Wrapper..get_points()\n')

        point_url = self.base_url + '/DkbTransactionBanking/content/FavorableWorld/Overview.xhtml?$event=init'
        self.dkb_br.open(point_url)

        p_dic = {}
        soup = self.dkb_br.get_current_page()
        table = soup.find('table', attrs={'class': 'expandableTable'})
        if table:
            tbody = table.find('tbody')
            row = tbody.find_all('tr')[0]
            cols = row.find_all("td")
            # points
            points = re.sub(' +', ' ', cols[1].text.strip())
            points = points.replace('\n', '')
            (pointsamount, pointsex) = points.replace('  ', ' ').split(' ', 1)
            pointsamount = int(pointsamount.replace('.', ''))
            pointsex = int(pointsex.replace('.', ''))

            # text
            ptext = cols[0].text.strip()
            (tlist) = ptext.split('\n', 1)
            ptext = tlist[0].lstrip()
            ptext = ptext.rstrip()
            etext = tlist[1].lstrip()
            etext = etext.rstrip()

            # store data
            p_dic[ptext] = pointsamount
            p_dic[etext] = pointsex

        logger.debug('Wrapper..get_points() ended\n')
        return p_dic

    def get_standing_orders(self, _uid: str = None) -> List[str]:
        """ get standing orders  """
        logger.debug('Wrapper.get_standing_orders()\n')

        so_url = self.base_url + '/ssohl/banking/finanzstatus/dauerauftraege?$event=infoStandingOrder'
        self.dkb_br.open(so_url)

        so_list = []
        soup = self.dkb_br.get_current_page()
        table = soup.find('table', attrs={'class': 'expandableTable'})
        if table:
            tbody = table.find('tbody')
            rows = tbody.find_all('tr')
            for row in rows:
                tmp_dic = {}
                cols = row.find_all("td")
                tmp_dic['recipient'] = cols[0].text.strip()
                amount = cols[1].text.strip()
                amount = amount.replace('\n', '')
                amount = string2float(amount.replace('EUR', ''))
                tmp_dic['amount'] = amount

                interval = cols[2]
                for brt in interval.find_all('br'):
                    brt.unwrap()

                interval = re.sub('\t+', ' ', interval.text.strip())
                interval = interval.replace('\n', '')
                interval = re.sub(' +', ' ', interval)
                tmp_dic['interval'] = interval
                tmp_dic['purpose'] = cols[3].text.strip()

                # store dict in list
                so_list.append(tmp_dic)

        logger.debug('Wrapper.get_standing_orders() ended\n')
        return so_list

    def get_transactions(self, transaction_url: str, atype: str, date_from: str, date_to: str, transaction_type: str = 'booked') -> List[str]:
        """ get transactions for a certain amount of time       """
        logger.debug('Wrapper.get_transactions(%s/%s: %s/%s, %s)\n', transaction_url, atype, date_from, date_to, transaction_type)

        transaction_list = []
        if atype == 'account':
            transaction_list = self._get_account_transactions(transaction_url, date_from, date_to, transaction_type)
        elif atype == 'creditcard':
            transaction_list = self._get_creditcard_transactions(transaction_url, date_from, date_to, transaction_type)
        elif atype == 'depot':
            transaction_list = self._get_depot_status(transaction_url, date_from, date_to, transaction_type)

        logger.debug('Wrapper.get_transactions() ended\n')
        return transaction_list

    def login(self) -> Tuple[Dict[str, str], str]:
        """ login into DKB banking area via the legacy frontend """
        logger.debug('Wrapper.login()\n')

        # login url
        login_url = self.base_url + '/ssohl/' + 'banking'

        # create browser and login
        self.dkb_br = self._new_instance()

        last_login = None
        account_dic = {}

        self.dkb_br.open(login_url)
        try:
            self.dkb_br.select_form('#login')
            self.dkb_br["j_username"] = str(self.dkb_user)
            self.dkb_br["j_password"] = str(self.dkb_password)

            # submit form and check response
            self.dkb_br.submit_selected()
            soup = self.dkb_br.get_current_page()

            # catch login error
            if soup.find("div", attrs={'class': 'clearfix module text errorMessage'}):
                raise DKBRoboError('Login failed')

            # catch generic notices
            if soup.find("form", attrs={'id': 'genericNoticeForm'}):
                self.dkb_br.open(login_url)
                soup = self.dkb_br.get_current_page()

            # filter last login date
            if soup.find("div", attrs={'id': 'lastLoginContainer'}):
                last_login = soup.find("div", attrs={'id': 'lastLoginContainer'}).text.strip()
                # remove crlf
                last_login = last_login.replace('\n', '')
                # format string in a way we need it
                last_login = last_login.replace('  ', '')
                last_login = last_login.replace('Letzte Anmeldung:', '')
                if soup.find('h1').text.strip() == 'Anmeldung bestätigen':
                    if self.tan_insert:
                        # chiptan input
                        login_confirmed = self._ctan_check(soup)
                    else:
                        # app confirmation needed to continue
                        login_confirmed = self._login_confirm()
                    if login_confirmed:
                        # login got confirmed get overview and parse data
                        soup_new = self._get_financial_statement()
                        account_dic = self._parse_overview(soup_new)
        except mechanicalsoup.utils.LinkNotFoundError as err:
            raise DKBRoboError('Login failed: LinkNotFoundError') from err

        logger.debug('Wrapper.login() ended\n')
        return account_dic, last_login

    def logout(self):
        """ logout from DKB banking area """
        logger.debug('Wrapper.logout()\n')

        logout_url = self.base_url + '/ssohl/' + 'DkbTransactionBanking/banner.xhtml?$event=logout'
        if self.dkb_br:
            self.dkb_br.open(logout_url)
        logger.debug('Wrapper.logout() ended\n')

    def scan_postbox(self, path: str = None, download_all: bool = False, archive: bool = False, prepend_date: bool = False):
        """ scans the DKB postbox and creates a dictionary """
        logger.debug('Wrapper.san_postbox() path: %s, download_all: %s, archive: %s, prepend_date: %s\n', path, download_all, archive, prepend_date)

        if archive:
            pb_url = self.base_url + '/ssohl/banking/postfach/ordner?$event=gotoFolder&folderNameOrId=archiv'
        else:
            pb_url = self.base_url + '/ssohl/banking/postfach'
        self.dkb_br.open(pb_url)
        soup = self.dkb_br.get_current_page()
        if archive:
            table = soup.find('table', attrs={'id': re.compile('mbo-message-list*')})
        else:
            table = soup.find('table', attrs={'id': 'welcomeMboTable'})
        if table is None:
            raise DKBRoboError("Expected table not found in old postbox")
        tbody = table.find('tbody')

        if archive:
            select_all = True
        else:
            select_all = download_all

        pb_dic = {}
        for row in tbody.find_all('tr'):
            link = row.find('a')
            link_name = link.contents[0]
            pb_dic[link_name] = {}
            pb_dic[link_name]['name'] = link_name
            pb_dic[link_name]['details'] = self.base_url + link['href']
            if path:
                pb_dic[link_name]['documents'] = self._get_document_links(pb_dic[link_name]['details'], path, link_name, select_all, prepend_date)
            else:
                pb_dic[link_name]['documents'] = self._get_document_links(pb_dic[link_name]['details'], select_all=select_all, prepend_date=prepend_date)

        logger.debug('Wrapper.scan_postbox() ended\n')
        return pb_dic



# pylint: disable=c3001, e1101, r0913, w0108, w0622
""" dkb_robo cli """
from datetime import date
from pathlib import Path
import pathlib
from pprint import pprint
import sys
import csv
import json
import tabulate
import click
import dkb_robo
from dkb_robo.utilities import object2dictionary
sys.path.append("..")

DATE_FORMAT = "%d.%m.%Y"
DATE_FORMAT_ALTERNATE = "%Y-%m-%d"


def _account_lookup(ctx, name, account, account_dic, unfiltered):
    """ lookup account """

    mapping_matrix = {
        'account': 'iban',
        'creditCard': 'maskedPan',
        'debitCard': 'maskedPan',
        'depot': 'depositAccountId',
        'brokerageAccount': 'depositAccountId',

    }
    if name is not None and account is None:
        if unfiltered:
            def account_filter(acct):
                product = getattr(acct, 'product', None)
                return getattr(product, 'displayName', None) == name
        else:
            def account_filter(acct):
                return acct["name"] == name
    elif account is not None and name is None:
        if unfiltered:
            def account_filter(acct):
                return getattr(acct, mapping_matrix[acct.type]) == account
        else:
            def account_filter(acct):
                return acct["account"] == account
    else:
        raise click.UsageError("One of --name or --account must be provided.", ctx)

    filtered_accounts = [
        acct for acct in account_dic.values() if account_filter(acct)
    ]
    if len(filtered_accounts) == 0:
        click.echo(f"No account found matching '{name or account}'", err=True)
        raise click.Abort()

    return filtered_accounts[0]


def _id_lookup(ctx, name, account, account_dic, unfiltered):
    """ lookup id """
    the_account = _account_lookup(ctx, name, account, account_dic, unfiltered)
    if unfiltered:
        uid = getattr(the_account, 'id', None)
    else:
        uid = the_account['id']
    return uid


def _transactionlink_lookup(ctx, name, account, account_dic, unfiltered):
    """ lookup id """
    the_account = _account_lookup(ctx, name, account, account_dic, unfiltered)

    if unfiltered:
        output_dic = {
            'id': getattr(the_account, 'id', None),
            'type': getattr(the_account, 'type', None),
            'transactions': getattr(the_account, 'transactions', None)
        }
    else:
        output_dic = {
            'id': the_account.get('id', None),
            'type': the_account.get('type', None),
            'transactions': the_account.get('transactions', None)
        }
    return output_dic


@click.group()
@click.option(
    "--debug",
    "-d",
    default=False,
    help="Show additional debugging",
    is_flag=True,
    envvar="DKB_DEBUG",
)
@click.option(
    "--unfiltered",
    default=False,
    is_flag=True,
    envvar="DKB_UNFILTERED",
    help="Do not filter output from DKB API",
)
@click.option(
    "--mfa-device",
    "-m",
    default=None,
    help='MFA device used for login ("1", "2" ...)',
    type=int,
    envvar="MFA_DEVICE",
)
@click.option(
    "--use-tan",
    default=False,
    is_flag=True,
    envvar="DKB_USE_TAN",
    hidden=True,
)
@click.option(
    "--chip-tan",
    "-t",
    default=None,
    help='use ChipTAN for login (either "qr" or "manual")',
    type=str,
    envvar="DKB_CHIP_TAN",
)
@click.option(
    "--username",
    "-u",
    required=True,
    type=str,
    help="username to access the dkb portal",
    envvar="DKB_USERNAME",
)
@click.option(
    "--password",
    "-p",
    prompt=True,
    hide_input=True,
    type=str,
    help="corresponding login password",
    envvar="DKB_PASSWORD",
)
@click.option(
    "--format",
    default="pprint",
    type=click.Choice(["pprint", "table", "csv", "json"]),
    help="output format to use",
    envvar="DKB_FORMAT",
)
@click.pass_context
def main(ctx, debug, unfiltered, mfa_device, use_tan, chip_tan, username, password, format):  # pragma: no cover
    """ main fuunction """

    if use_tan:
        click.echo("The --use-tan option is deprecated and will be removed in a future release. Please use the --chip-tan option", err=True)
        chip_tan = True
    ctx.ensure_object(dict)
    ctx.obj["DEBUG"] = debug
    ctx.obj["UNFILTERED"] = unfiltered
    ctx.obj["CHIP_TAN"] = chip_tan
    ctx.obj["MFA_DEVICE"] = mfa_device
    ctx.obj["USERNAME"] = username
    ctx.obj["PASSWORD"] = password
    ctx.obj["FORMAT"] = _load_format(format)


@main.command()
@click.pass_context
def accounts(ctx):
    """ get list of account """
    try:
        with _login(ctx) as dkb:
            accounts_dict = dkb.account_dic
            for id, value in accounts_dict.items():
                if ctx.obj["UNFILTERED"]:
                    value = object2dictionary(value)
                    accounts_dict[id] = value
                if 'details' in value:
                    del value["details"]
                if 'transactions' in value:
                    del value["transactions"]
            ctx.obj["FORMAT"](list(accounts_dict.values()))
    except dkb_robo.DKBRoboError as _err:
        click.echo(_err.args[0], err=True)


@main.command()
@click.pass_context
@click.option(
    "--name",
    "-n",
    type=str,
    help="Name of the account to fetch transactions for",
    envvar="DKB_TRANSACTIONS_ACCOUNT_NAME",
)
@click.option(
    "--account",
    "-a",
    type=str,
    help="Account to fetch transactions for",
    envvar="DKB_TRANSACTIONS_ACCOUNT",
)
@click.option(
    "--transaction-type",
    "-t",
    default="booked",
    type=click.Choice(["booked", "reserved"]),
    help="The type of transactions to fetch",
    envvar="DKB_TRANSACTIONS_TYPE",
)
@click.option(
    "--date-from",
    type=click.DateTime(formats=[DATE_FORMAT, DATE_FORMAT_ALTERNATE]),
    default=date.today().strftime(DATE_FORMAT),
)
@click.option(
    "--date-to",
    type=click.DateTime(formats=[DATE_FORMAT, DATE_FORMAT_ALTERNATE]),
    default=date.today().strftime(DATE_FORMAT),
)
def transactions(ctx, name, account, transaction_type, date_from, date_to):  # pragma: no cover
    """ get list of transactions """

    try:
        with _login(ctx) as dkb:
            the_account = _transactionlink_lookup(ctx, name, account, dkb.account_dic, ctx.obj["UNFILTERED"])
            transactions_list = dkb.get_transactions(
                the_account["transactions"],
                the_account["type"],
                date_from.strftime(DATE_FORMAT),
                date_to.strftime(DATE_FORMAT),
                transaction_type=transaction_type,
            )
            ctx.obj["FORMAT"](transactions_list)

    except dkb_robo.DKBRoboError as _err:
        click.echo(_err.args[0], err=True)


@main.command()
@click.pass_context
def last_login(ctx):
    """ get last login """
    try:
        with _login(ctx) as dkb:
            ctx.obj["FORMAT"]([{"last_login": dkb.last_login}])
    except dkb_robo.DKBRoboError as _err:
        click.echo(_err.args[0], err=True)


@main.command()
@click.pass_context
def credit_limits(ctx):
    """ get limits """
    try:
        with _login(ctx) as dkb:
            limits = dkb.get_credit_limits()
            limits = [{"account": k, "limit": v} for k, v in limits.items()]
            ctx.obj["FORMAT"](limits)
    except dkb_robo.DKBRoboError as _err:
        click.echo(_err.args[0], err=True)


@main.command()
@click.pass_context
@click.option(
    "--name",
    "-n",
    type=str,
    help="Name of the account to fetch transactions for",
    envvar="DKB_TRANSACTIONS_ACCOUNT_NAME",
)
@click.option(
    "--account",
    "-a",
    type=str,
    help="Account to fetch transactions for",
    envvar="DKB_TRANSACTIONS_ACCOUNT",
)
def standing_orders(ctx, name, account):  # pragma: no cover
    """ get standing orders """
    try:
        with _login(ctx) as dkb:
            uid = _id_lookup(ctx, name, account, dkb.account_dic, ctx.obj["UNFILTERED"])
            so_list = dkb.get_standing_orders(uid)
            standing_orders_list = []
            for so in so_list:
                if ctx.obj["UNFILTERED"]:
                    standing_orders_list.append(object2dictionary(so))
                else:
                    standing_orders_list.append(so)
            ctx.obj["FORMAT"](standing_orders_list)
    except dkb_robo.DKBRoboError as _err:
        click.echo(_err.args[0], err=True)


@main.command()
@click.pass_context
@click.option(
    "--path",
    "-p",
    type=str,
    help="Path to save the documents to",
    envvar="DKB_DOC_PATH",
)
@click.option("--download_all", is_flag=True, show_default=True, default=False, help="Download all documents", envvar="DKB_DOWNLOAD_ALL")
@click.option("--archive", is_flag=True, show_default=True, default=False, help="Download archive", envvar="DKB_ARCHIVE")
@click.option("--prepend_date", is_flag=True, show_default=True, default=False, help="Prepend date to filename", envvar="DKB_PREPEND_DATE")
def scan_postbox(ctx, path, download_all, archive, prepend_date):
    """ scan postbox """
    if not path:
        path = "documents"
    try:
        with _login(ctx) as dkb:
            doc_list = dkb.scan_postbox(path=path, download_all=download_all, prepend_date=prepend_date)
            if ctx.obj["UNFILTERED"]:
                documents_list = []
                for doc in doc_list.values():
                    documents_list.append(object2dictionary(doc))
            else:
                documents_list = doc_list
            ctx.obj["FORMAT"](documents_list)
    except dkb_robo.DKBRoboError as _err:
        click.echo(_err.args[0], err=True)


@main.command()
@click.pass_context
@click.option(
    "--path",
    "-p",
    type=click.Path(writable=True, path_type=pathlib.Path),
    help="Path to save the documents to",
    envvar="DKB_DOC_PATH",
)
@click.option("--all", "-A", is_flag=True, show_default=True, default=False, help="Download all documents", envvar="DKB_DOWNLOAD_ALL")
@click.option("--prepend-date", is_flag=True, show_default=True, default=False, help="Prepend date to filename", envvar="DKB_PREPEND_DATE")
@click.option("--mark-read", is_flag=True, show_default=True, default=True, help="Mark downloaded files read", envvar="DKB_MARK_READ")
@click.option("--use-account-folders", is_flag=True, show_default=True, default=False, help="Store files in separate folders per account/depot", envvar="DKB_ACCOUNT_FOLDERS")
@click.option("--list-only", is_flag=True, show_default=True, default=False, help="Only list documents, do not download", envvar="DKB_LIST_ONLY")
def download(ctx, path: Path, all: bool, prepend_date: bool, mark_read: bool, use_account_folders: bool, list_only: bool):
    """ download document """
    if path is None:
        list_only = True
    try:
        with _login(ctx) as dkb:
            ctx.obj["FORMAT"](dkb.download(path=path, download_all=all, prepend_date=prepend_date, mark_read=mark_read, use_account_folders=use_account_folders, list_only=list_only))
    except dkb_robo.DKBRoboError as _err:
        click.echo(_err.args[0], err=True)


def _load_format(output_format):
    """ select output format based on cli option """
    if output_format == "pprint":
        return lambda data: pprint(data)

    if output_format == "table":
        return lambda data: click.echo(
            tabulate.tabulate(data, headers="keys", tablefmt="grid")
        )

    if output_format == "csv":
        def formatter(data):  # pragma: no cover
            if len(data) == 0:
                return
            writer = csv.DictWriter(sys.stdout, fieldnames=max(data, key=len).keys())
            writer.writeheader()
            writer.writerows(data)

        return formatter

    if output_format == "json":

        return lambda data: click.echo(json.dumps(data, indent=2))

    raise ValueError(f"Unknown format: {output_format}")


def _login(ctx):
    return dkb_robo.DKBRobo(
        dkb_user=ctx.obj["USERNAME"], dkb_password=ctx.obj["PASSWORD"], chip_tan=ctx.obj["CHIP_TAN"], debug=ctx.obj["DEBUG"], unfiltered=ctx.obj["UNFILTERED"], mfa_device=ctx.obj["MFA_DEVICE"]
    )



""" Module for handling dkb transactions """
# pylint: disable=c0415, r0913, c0103
import datetime
import time
from typing import Dict, List, Optional, Union
from dataclasses import dataclass, field
import logging
import requests
from dkb_robo.utilities import Account, Amount, PerformanceValue, get_dateformat, filter_unexpected_fields, ulal


LEGACY_DATE_FORMAT, API_DATE_FORMAT = get_dateformat()
logger = logging.getLogger(__name__)


class Transactions:
    """ Transactions class """

    def __init__(self, client: requests.Session, unfiltered: bool = False, base_url: str = 'https://banking.dkb.de/api'):
        self.client = client
        self.base_url = base_url
        self.uid = None
        self.unfiltered = unfiltered

    def _correlate(self, transaction_dic: Dict[str, str]) -> List[Dict[str, str]]:
        """ correlate transactions """
        logger.debug('Transactions._correlate()\n')

        if 'included' in transaction_dic:
            included_list = transaction_dic['included']
        else:
            included_list = []

        position_list = []
        if 'data' in transaction_dic:
            for position in transaction_dic['data']:
                position_dic = self._map(position, included_list)
                if position_dic:
                    position_list.append(position_dic)

        logger.debug('Transactions._correlate() ended with %s entries\n', len(position_list))
        return position_list

    def _fetch(self, transaction_url: str) -> Dict[str, str]:
        """ get transaction list"""
        logger.debug('Transactions.fetch(%s)\n', transaction_url)

        transaction_dic = {'data': [], 'included': []}
        while transaction_url:
            response = self.client.get(transaction_url)
            if response.status_code == 200:
                _transaction_dic = response.json()
                if 'data' in _transaction_dic:
                    transaction_dic['data'].extend(_transaction_dic['data'])
                    transaction_url = self._nextpage_url(_transaction_dic)    # get next page
                else:
                    logger.debug('fetch transactions: no data in response')
                    transaction_url = None

                if 'included' in _transaction_dic:
                    transaction_dic['included'].extend(_transaction_dic['included'])
            else:
                logger.error('fetch transactions: http status code is not 200 but %s', response.status_code)
                break

        logger.debug('Transactions.fetch() ended with %s entries\n', len(transaction_dic['data']))
        return transaction_dic

    def _filter(self, transaction_list: List[Dict[str, str]], date_from: str, date_to: str, transaction_type: str) -> List[Dict[str, str]]:
        """ filter transactions """
        logger.debug('Transactions._filter()\n')

        # support transation type 'reserved' for backwards compatibility
        transaction_type = 'pending' if transaction_type == 'reserved' else transaction_type

        try:
            date_from_uts = int(time.mktime(datetime.datetime.strptime(date_from, LEGACY_DATE_FORMAT).timetuple()))
        except ValueError:
            date_from_uts = int(time.mktime(datetime.datetime.strptime(date_from, API_DATE_FORMAT).timetuple()))

        try:
            date_to_uts = int(time.mktime(datetime.datetime.strptime(date_to, LEGACY_DATE_FORMAT).timetuple()))
        except ValueError:
            date_to_uts = int(time.mktime(datetime.datetime.strptime(date_to, API_DATE_FORMAT).timetuple()))

        filtered_transaction_list = []
        for transaction in transaction_list:
            if 'attributes' in transaction and 'status' in transaction['attributes'] and 'bookingDate' in transaction['attributes']:
                if transaction['attributes']['status'] == transaction_type:
                    bookingdate_uts = int(time.mktime(datetime.datetime.strptime(transaction['attributes']['bookingDate'], API_DATE_FORMAT).timetuple()))
                    if date_from_uts <= bookingdate_uts <= date_to_uts:
                        filtered_transaction_list.append(transaction)

        logger.debug('Transactions._filter() ended with %s entries\n', len(filtered_transaction_list))
        return filtered_transaction_list

    def _format(self, raw_transaction_list: List[Dict[str, str]], atype: str) -> List[Dict[str, str]]:
        """ format transaction list """
        logger.debug('Transactions._format()\n')

        mapping_dic = {
            'account': 'AccountTransactionItem',
            'creditcard': 'CreditCardTransactionItem',
            'creditCard': 'CreditCardTransactionItem',
            'brokerageAccount': 'DepotTransactionItem',
            'depot': 'DepotTransactionItem'
        }

        transaction_list = []
        if raw_transaction_list:
            for ele in raw_transaction_list:
                if 'attributes' not in ele or 'id' not in ele:
                    continue
                # add id to attributes tree
                ele['attributes']['id'] = ele['id']
                transaction = globals()[mapping_dic[atype]](**ele['attributes'])
                if self.unfiltered:
                    transaction_list.append(transaction)
                else:
                    transaction_list.append(transaction.format())

        logger.debug('Transactions._format() ended with %s entries\n', len(transaction_list))
        return transaction_list

    def _map(self, position: Dict[str, str], included_list: List[Dict[str, str]]):
        """ add details from depot transaction """
        logger.debug('DepotTransaction._map()\n')

        instrument_id = position.get('relationships', {}).get('instrument', {}).get('data', {}).get('id', None)
        quote_id = position.get('relationships', {}).get('quote', {}).get('data', {}).get('id', None)
        for ele in included_list:
            if 'id' in ele and ele['id'] == instrument_id:
                position['attributes']['instrument'] = ele['attributes']
                position['attributes']['instrument']['id'] = ele['id']
            if 'id' in ele and ele['id'] == quote_id:
                position['attributes']['quote'] = ele['attributes']
                position['attributes']['quote']['id'] = ele['id']

        logger.debug('DepotTransaction._map() ended\n')
        return position

    def _nextpage_url(self, tr_dic):
        """ get transaction url """
        logger.debug('Transactions._nextpage_url()\n')

        transaction_url = None
        if 'links' in tr_dic and 'next' in tr_dic['links']:
            logger.debug('Transactions._nextpage_url(): next page: %s', tr_dic['links']['next'])
            transaction_url = self.base_url + '/accounts' + tr_dic['links']['next']
        else:
            logger.debug('Transactions._nextpage_url(): no next page')
            transaction_url = None

        logger.debug('Transactions._nextpage_url() ended\n')
        return transaction_url

    def get(self, transaction_url: str, atype: str, date_from: str, date_to: str, transaction_type: str = 'booked'):
        """ fetch transactions """
        logger.debug('Transactions.get()\n')

        if transaction_url:
            if atype == 'account':
                logger.info('fetching account transactions')
                transaction_url = transaction_url + '?filter[bookingDate][GE]=' + date_from + '&filter[bookingDate][LE]=' + date_to + '&expand=Merchant&page[size]=400'
            elif atype in ['creditcard', 'creditCard']:
                logger.info('fetching card transactions')
                transaction_url = transaction_url + '&filter[date][GE]=' + date_from + '&filter[date][LE]=' + date_to + '&expand=Merchant&page[size]=400'
        transaction_dic = self._fetch(transaction_url)

        if atype in ['account', 'creditcard', 'creditCard']:
            raw_transaction_list = self._filter(transaction_list=transaction_dic['data'], date_from=date_from, date_to=date_to, transaction_type=transaction_type)
        else:
            raw_transaction_list = self._correlate(transaction_dic)

        # format output
        transaction_list = self._format(raw_transaction_list, atype)

        logger.debug('Transactions.get() ended\n')
        return transaction_list


@filter_unexpected_fields
@dataclass
class AccountTransactionItem:
    """ dataclass for a single AccountTransaction """
    status: Optional[str] = None
    bookingDate: Optional[str] = None
    valueDate: Optional[str] = None
    description: Optional[str] = None
    mandateId: Optional[str] = None
    endToEndId: Optional[str] = None
    transactionType: Optional[str] = None
    purposeCode: Optional[str] = None
    businessTransactionCode: Optional[str] = None
    amount: Optional[Dict] = None
    creditor: Optional[Union[Dict, str]] = None
    debtor: Optional[Union[Dict, str]] = None
    isRevocable: bool = False

    def __post_init__(self):
        self.amount = ulal(Amount, self.amount)
        # regroup creditor information allowing simpler access
        self.creditor = ulal(Account, self._peer_information(self.creditor, 'creditorAccount'))
        # regroup debtor for the same reason
        self.debtor = ulal(Account, self._peer_information(self.debtor, 'debtorAccount'))
        if self.description:
            self.description = " ".join(self.description.split())

    def _peer_information(self, peer_dic: Dict[str, str], peer_type: str = None) -> Dict[str, str]:
        """ add peer information """
        logger.debug('AccountTransaction._peer_information(%s)\n', peer_type)

        peer_dic[peer_type]['bic'] = peer_dic.get('agent', {}).get('bic', None)
        peer_dic[peer_type]['id'] = peer_dic.get('id', None)

        try:
            peer_dic[peer_type]['name'] = " ".join(peer_dic.pop('name', None).split())
        except Exception:
            peer_dic[peer_type]['name'] = peer_dic.pop('name', None)

        if peer_dic.get('intermediaryName', None):
            peer_dic[peer_type]['intermediaryName'] = " ".join(peer_dic.get('intermediaryName', None).split())

        logger.debug('AccountTransaction._peer_information() ended\n')
        return peer_dic[peer_type]

    def format(self):
        """ format format transaction list ot a useful output """
        logger.debug('AccountTransaction.format()\n')

        transaction_dic = {
            'amount': self.amount.value,
            'currencycode': self.amount.currencyCode,
            'date': self.bookingDate,
            # for backwards compatibility
            'bdate': self.bookingDate,
            'vdate': self.valueDate,
            'customerreference': self.endToEndId,
            'mandatereference': self.mandateId,
            'postingtext': self.transactionType,
            'reasonforpayment': self.description
        }

        if self.amount.value > 0:
            # incoming transaction
            transaction_dic['peeraccount'] = self.debtor.iban
            transaction_dic['peerbic'] = self.debtor.bic
            transaction_dic['peerid'] = self.debtor.id
            if self.debtor.intermediaryName:
                transaction_dic['peer'] = self.debtor.intermediaryName
            else:
                transaction_dic['peer'] = self.debtor.name
        else:
            # outgoing transaction
            transaction_dic['peeraccount'] = self.creditor.iban
            transaction_dic['peerbic'] = self.creditor.bic
            transaction_dic['peerid'] = self.creditor.id
            if self.creditor.intermediaryName and self.description.lower() != 'visa debitkartenumsatz':
                transaction_dic['peer'] = self.creditor.intermediaryName
            else:
                transaction_dic['peer'] = self.creditor.name

        # this is for backwards compatibility
        transaction_dic['text'] = f'{transaction_dic["postingtext"]} {transaction_dic["peer"]} {transaction_dic["reasonforpayment"]}'

        logger.debug('AccountTransaction.format() ended\n')
        return transaction_dic


@filter_unexpected_fields
@dataclass
class CreditCardTransactionItem:
    """ dataclass for a single CreditCardTransaction """
    amount: Optional[Dict] = None
    id: Optional[str] = None
    authorizationDate: Optional[str] = None
    bonuses: Optional[List] = None
    bookingDate: Optional[str] = None
    cardId: Optional[str] = None
    description: Optional[str] = None
    merchantAmount: Optional[Dict] = None
    merchantCategory: Optional[Dict] = field(default_factory=dict)
    status: Optional[str] = None
    transactionType: Optional[str] = None

    def __post_init__(self):
        self.amount = ulal(Amount, self.amount)
        self.merchantAmount = ulal(Amount, self.merchantAmount)
        self.merchantCategory = ulal(self.MerchantCategory, self.merchantCategory)

    @filter_unexpected_fields
    @dataclass
    class MerchantCategory:
        """ dataclass for a single merchantCategory """
        code: Optional[str] = None

    def format(self):
        """ format format transaction list ot a useful output """
        logger.debug('CreditCardTransaction.format()\n')

        transaction_dic = {
            # fixing strange behaviour of DKB API
            'amount': self.amount.value,  # * -1,
            'bdate': self.bookingDate,
            'currencycode': self.amount.currencyCode,
            'text': self.description,
            'vdate': self.authorizationDate,
        }

        logger.debug('CreditCardTransaction.format() ended\n')
        return transaction_dic


@filter_unexpected_fields
@dataclass
class DepotTransactionItem:
    """ DepotTransaction class """

    id: Optional[str] = None
    availableQuantity: Optional[Union[Dict, str]] = None
    custody: Optional[Union[Dict, str]] = None
    instrument: Optional[Union[Dict, str]] = None
    lastOrderDate: Optional[str] = None
    performance: Optional[Union[Dict, str]] = None
    quantity: Optional[Union[Dict, str]] = None
    quote: Optional[Union[Dict, str]] = None

    def __post_init__(self):
        self.availableQuantity = ulal(self.Quantity, self.availableQuantity)
        self.custody = ulal(self.Custody, self.custody)
        self.performance = ulal(self.Performance, self.performance)
        self.quantity = ulal(self.Quantity, self.quantity)
        self.instrument = ulal(self.Instrument, self.instrument)
        self.quote = ulal(self.Quote, self.quote)

    @filter_unexpected_fields
    @dataclass
    class Custody:
        """ dataclass for custody """
        block: Optional[Union[Dict, str]] = None
        certificateType: Optional[str] = None
        characteristic: Optional[Union[Dict, str]] = None
        custodyType: Optional[str] = None
        custodyTypeId: Optional[str] = None

        @filter_unexpected_fields
        @dataclass
        class Block:
            """ dataclass for block """
            blockType: Optional[str] = None

        @filter_unexpected_fields
        @dataclass
        class Characteristic:
            """ dataclass for characteristic """
            characteristicType: Optional[str] = None

        def __post_init__(self):
            self.block = ulal(self.Block, self.block)
            self.characteristic = ulal(self.Characteristic, self.characteristic)

    @filter_unexpected_fields
    @dataclass
    class Instrument:
        """ dataclass for instrument """
        id: Optional[str] = None
        identifiers: Optional[List] = field(default_factory=list)
        name: Optional[Union[Dict, str]] = None
        unit: Optional[str] = None

        def __post_init__(self):
            self.name = ulal(self.Name, self.name)
            self.identifiers = [self.IdentifierItem(**identifier) for identifier in self.identifiers]

        @filter_unexpected_fields
        @dataclass
        class IdentifierItem:
            """ dataclass for identifier """
            identifier: Optional[str] = None
            value: Optional[str] = None

        @filter_unexpected_fields
        @dataclass
        class Name:
            """ dataclass for name """
            long: Optional[str] = None
            short: Optional[str] = None

    @filter_unexpected_fields
    @dataclass
    class Performance:
        """ dataclass for performance """
        currentValue: Optional[Union[Dict, str]] = None
        isOutdated: Optional[bool] = False

        def __post_init__(self):
            self.currentValue = ulal(PerformanceValue, self.currentValue)

    @filter_unexpected_fields
    @dataclass
    class Quantity:
        """ dataclass for quantity """
        unit: Optional[str] = None
        value: Optional[float] = None

        def __post_init__(self):
            try:
                self.value = float(self.value)
            except Exception:
                self.value = None

    @filter_unexpected_fields
    @dataclass
    class Quote:
        """ dataclass for quote """
        id: Optional[str] = None
        market: Optional[str] = None
        price: Optional[Union[Dict, str]] = None
        timestamp: Optional[str] = None

        def __post_init__(self):
            self.price = ulal(PerformanceValue, self.price)

    def format(self) -> Dict[str, str]:
        """ format  transaction list ot a useful output """
        logger.debug('DepotTransaction.format()\n')

        transaction_dic = {
            'isin_wkn': self.instrument.identifiers[0].value,
            'lastorderdate': self.lastOrderDate,
            'price_euro': self.performance.currentValue.value,
            'quantity': self.quantity.value,
            # for backwards compatibility
            'shares': self.availableQuantity.value,
            'shares_unit': self.quantity.unit,
            'text': self.instrument.name.short,
            'text_long': self.instrument.name.long,
        }

        if self.quote:
            transaction_dic['currencyCode'] = self.quote.price.currencyCode
            transaction_dic['market'] = self.quote.market
            transaction_dic['price'] = self.quote.price.value

        return transaction_dic
